# -*- coding: utf-8 -*-
# ####################################################################
#  Copyright (C) 2005-2019 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

""" Savers plugin manager  """
from __future__ import print_function

import os.path

from fife import fife
from fife.extensions.serializers.xmlmapsaver import XMLMapSaver

mapFileMapping = { 'xml' : XMLMapSaver}
fileExtensions = ('xml',)

def saveMapFile(path, engine, map, importList=[], debug=True):
	""" save map file
	@type	path:		string
	@param	path:		The fully qualified path to the file to save
	@type	engine:		object
	@param	engine: 	FIFE engine instance
	@type	map:		object
	@param	map:		FIFE map object
	@type	importList:	list
	@param	importList:	A list of all imports
	@type 	debug:		boolean
	@param	debug:		Enables debugging information
	"""
	(filename, extension) = os.path.splitext(path)
	map.setFilename(path)
	map_saver = mapFileMapping[extension[1:]](path, engine, map, importList)

	map_saver.saveResource()
	if debug: print("--- Saved Map.")
	return map

def addMapSaver(fileExtension, saverClass):
	"""Add a new saver for fileextension
	@type   fileExtension: string
	@param  fileExtension: The file extension the saver is registered for
	@type   saverClass:   object
	@param  saverClass:   A fife.ResourceLoader implementation that saves maps
	                       from files with the given fileExtension
	"""
	mapFileMapping[fileExtension] = saverClass
	_updateMapFileExtensions()

def _updateMapFileExtensions():
	global fileExtensions
	fileExtensions = set(mapFileMapping.keys())



# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2019 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

from __future__ import absolute_import
from fife.extensions.pychan.attrs import UnicodeAttr

from .common import gui2text, text2gui
from .widget import Widget


class BasicTextWidget(Widget):
	"""
	The base class for widgets which display a string - L{Label},L{Button}, etc.
	Do not use directly.

	New Attributes
	==============

	  - text: The text (depends on actual widget)

	Data
	====

	The text can be set via the L{distributeInitialData} method.
	"""

	ATTRIBUTES = Widget.ATTRIBUTES + [ UnicodeAttr('text') ]

	DEFAULT_HEXPAND = True
	DEFAULT_VEXPAND = False
	DEFAULT_MARGINS = 0,0
	DEFAULT_PADDING = 0
	
	DEFAULT_TEXT = u""
	
	def __init__(self, 
				 parent = None,
				 name = None,
				 size = None,
				 min_size = None,
				 max_size = None,
				 fixed_size = None,
				 margins = None,
				 padding = None,
				 helptext = None,
				 position = None,
				 style = None,
				 hexpand = None,
				 vexpand = None,
				 font = None,
				 base_color = None,
				 background_color = None,
				 foreground_color = None,
				 selection_color = None,
				 border_color = None,
				 outline_color = None,
				 border_size = None,
				 outline_size = None,
				 position_technique = None,
				 is_focusable = None,
				 comment = None,
				 text = None):
				 
		self.text = self.DEFAULT_TEXT
		super(BasicTextWidget,self).__init__(parent=parent,
											 name=name,
											 size=size,
											 min_size=min_size,
											 max_size=max_size,
											 fixed_size=fixed_size,
											 margins=margins,
											 padding=padding,
											 helptext=helptext,
											 position=position,
											 style=style,
											 hexpand=hexpand,
											 vexpand=vexpand,
											 font=font,
											 base_color=base_color,
											 background_color=background_color,
											 foreground_color=foreground_color,
											 selection_color=selection_color,
											 border_color=border_color,
											 outline_color=outline_color,
											 border_size=border_size,
											 outline_size=outline_size,
											 position_technique=position_technique,
											 is_focusable=is_focusable,
											 comment=comment)

		# Prepare Data collection framework
		self.accepts_initial_data = True
		self._realSetInitialData = self._setText
		
		# Override anything set when stylize was called
		if margins is not None: self.margins = margins
		if text is not None: self.text = text

	def _getText(self): return gui2text(self.real_widget.getCaption())
	def _setText(self,text): self.real_widget.setCaption(text2gui(text))

	text = property(_getText,_setText)



# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2019 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

"""
Pychan extension widgets.

Extension widgets are partly experimental, partly rarely used widgets
which are added here. They are by default not included in the widgets
registry and thus cannot be loaded from XML files. Use L{pychan.widgets.registerWidget}
to enable that.

Not the same care to keep the API stable will be taken for them and
before and if they are added (or replace) the standard widgets they
will have to be reviewed in detail.
"""



# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2013 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

from __future__ import absolute_import
from fife import fifechan

from .widget import Widget
from fife.extensions.pychan.attrs import PointAttr

class Spacer(Widget):
	"""
	A spacer represents expandable or fixed 'whitespace' in the GUI.

	In a XML file you can get this by adding a <Spacer /> inside a VBox or
	HBox element (Windows implicitly are VBox elements).
	"""
	
	DEFAULT_HEXPAND = True
	DEFAULT_VEXPAND = False
	DEFAULT_MARGINS = 0, 0
	DEFAULT_PADDING = 0
	
	def __init__(self, 
				 parent = None, 
				 name = None,
				 size = None,
				 min_size = None, 
				 max_size = None,
				 fixed_size = None,
				 margins = None,
				 padding = None,
				 helptext = None, 
				 position = None, 
				 style = None, 
				 hexpand = None,
				 vexpand = None,
				 font = None,
				 base_color = None,
				 background_color = None,
				 foreground_color = None,
				 selection_color = None,
				 border_color = None,
				 outline_color = None,
				 border_size = None,
				 outline_size = None,
				 position_technique = None,
				 is_focusable = None,
				 comment = None):
				 
		self.real_widget = fifechan.Spacer()
		super(Spacer,self).__init__(parent=parent,
									name=name,
									size=size,
									min_size=min_size,
									max_size=max_size,
									fixed_size=fixed_size,
									margins=margins,
									padding=padding,
									helptext=helptext,
									position=position,
									style=style,
									hexpand=hexpand,
									vexpand=vexpand,
									font=font,
									base_color=base_color,
									background_color=background_color,
									foreground_color=foreground_color,
									selection_color=selection_color,
									border_color=border_color,
									outline_color=outline_color,
									border_size=border_size,
									outline_size=outline_size,
									position_technique=position_technique,
									is_focusable=is_focusable,
									comment=comment)
		if hexpand is not None: 
			self.hexpand = hexpand
		else:
			self.hexpand = self.DEFAULT_HEXPAND
		if vexpand is not None: 
			self.vexpand = vexpand
		else:
			self.vexpand = self.DEFAULT_VEXPAND

	def clone(self, prefix):
		spacerClone = Spacer(None,
				self._createNameWithPrefix(prefix),
				self.size,
				self.min_size, 
				self.max_size,
				self.fixed_size,
				self.margins,
				self.padding,
				self.helptext, 
				self.position, 
				self.style, 
				self.hexpand,
				self.vexpand,
				self.font,
				self.base_color,
				self.background_color,
				self.foreground_color,
				self.selection_color,
				self.border_color,
				self.outline_color,
				self.border_size,
				self.outline_size,
				self.position_technique,
				self.is_focusable,
				self.comment)
				 
		return spacerClone
	



# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2019 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

from __future__ import absolute_import
from fife import fifechan

from fife.extensions.pychan.attrs import IntAttr

from .widget import Widget


class PercentageBar(Widget):
	""" A percentage bar widget

	New Attributes
	==============

	  - orientation: 1 = horizontal, 0=vertical
	  - value: int: default 0

	"""

	HORIZONTAL = fifechan.PercentageBar.HORIZONTAL
	VERTICAL = fifechan.PercentageBar.VERTICAL

	ATTRIBUTES = Widget.ATTRIBUTES + [ IntAttr('orientation'),
									   IntAttr('value')
									 ]
	DEFAULT_HEXPAND = True
	DEFAULT_VEXPAND = False
	DEFAULT_MIN_SIZE = 10,10
	DEFAULT_VALUE = 0
	DEFAULT_ORIENTATION = HORIZONTAL

	def __init__(self,
				 parent = None,
				 name = None,
				 size = None,
				 min_size = None,
				 max_size = None,
				 fixed_size = None,
				 margins = None,
				 padding = None,
				 helptext = None,
				 position = None,
				 style = None,
				 hexpand = None,
				 vexpand = None,
				 font = None,
				 base_color = None,
				 background_color = None,
				 foreground_color = None,
				 selection_color = None,
				 border_color = None,
				 outline_color = None,
				 border_size = None,
				 outline_size = None,
				 position_technique = None,
				 is_focusable = None,
				 comment = None,
				 value = None,
				 orientation = None):

		self.real_widget = fifechan.PercentageBar()
		self.orientation = self.DEFAULT_ORIENTATION
		self.value = self.DEFAULT_VALUE

		super(PercentageBar, self).__init__(parent=parent,
											name=name,
											size=size,
											min_size=min_size,
											max_size=max_size,
											fixed_size=fixed_size,
											margins=margins,
											padding=padding,
											helptext=helptext,
											position=position,
											style=style,
											hexpand=hexpand,
											vexpand=vexpand,
											font=font,
											base_color=base_color,
											background_color=background_color,
											foreground_color=foreground_color,
											selection_color=selection_color,
											border_color=border_color,
											outline_color=outline_color,
											border_size=border_size,
											outline_size=outline_size,
											position_technique=position_technique,
											is_focusable=is_focusable,
											comment=comment)

		if orientation is not None: self.orientation = orientation
		if value is not None: self.value = value

		self.accepts_data = True
		self._realSetData = self._setValue
		self._realGetData = self._getValue

	def clone(self, prefix):
		pbarClone = PercentageBar(None,
						self._createNameWithPrefix(prefix),
						self.size,
						self.min_size,
						self.max_size,
						self.fixed_size,
						self.margins,
						self.padding,
						self.helptext,
						self.position,
						self.style,
						self.hexpand,
						self.vexpand,
						self.font,
						self.base_color,
						self.background_color,
						self.foreground_color,
						self.selection_color,
						self.border_color,
						self.outline_color,
						self.border_size,
						self.outline_size,
						self.position_technique,
						self.is_focusable,
						self.comment,
						self.value,
						self.orientation)
		return pbarClone

	def _getValue(self):
		"""getValue(self) -> int"""
		return self.real_widget.getValue()

	def _setValue(self, value):
		"""setValue(self, int value)"""
		if type(value) != int:
			raise RuntimeError("PercentageBar only accepts integer values")
		self.real_widget.setValue(value)
	value = property(_getValue, _setValue)

	def _setOrientation(self, orientation):
		"""setOrientation(self, Orientation orientation)"""
		self.real_widget.setOrientation(orientation)

	def _getOrientation(self):
		"""getOrientation(self) -> int"""
		return self.real_widget.getOrientation()
	orientation = property(_getOrientation, _setOrientation)




# -*- coding: utf-8 -*-
# ####################################################################
#  Copyright (C) 2005-2019 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

""" utilities for xml maploading process """
from __future__ import print_function
from __future__ import division

from builtins import range
from past.utils import old_div
import os
import math

def loadImportFile(loader, path, engine, debug=False):
	""" uses XMLObjectLoader to load import files from path
	
	@type	path:	string
	@param	path:	path to import file
	@type	debug:	bool
	@param	debug:	flag to activate / deactivate print statements
	"""
	loader.loadResource(path)
	if debug: print('imported object file ' + path)

def loadImportDir(loader, path, engine, debug=False):
	""" helper function to call loadImportFile on a directory
	
	@type	path:	string
	@param	path:	path to import directory
	@type	debug:	bool
	@param	debug:	flag to activate / deactivate print statements
	"""
	for _file in [f for f in engine.getVFS().listFiles(path) if f.split('.')[-1] == 'xml']:
		loadImportFile(loader, '/'.join([path, _file]), engine, debug)

def loadImportDirRec(loader, path, engine, debug=False):
	""" helper function to call loadImportFile recursive on a directory

	@type	path:	string
	@param	path:	path to import directory
	@type	debug:	bool
	@param	debug:	flag to activate / deactivate print statements	
	"""
	loadImportDir(loader, path, engine, debug)

	for _dir in [d for d in engine.getVFS().listDirectories(path) if not d.startswith('.')]:
		loadImportDirRec(loader, '/'.join([path, _dir]), engine, debug)
		
def root_subfile(masterfile, subfile):
	"""
	Returns new path for given subfile (path), which is rooted against masterfile
	E.g. if masterfile is ./../foo/bar.xml and subfile is ./../foo2/subfoo.xml,
	returned path is ../foo2/subfoo.xml
	NOTE: masterfile is expected to be *file*, not directory. subfile can be either
	"""
	s = '/'

	masterfile = norm_path(os.path.abspath(masterfile))
	subfile = norm_path(os.path.abspath(subfile))

	master_fragments = masterfile.split(s)
	sub_fragments = subfile.split(s)

	master_leftovers = []
	sub_leftovers = []

	for i in range(len(master_fragments)):
		try:
			if master_fragments[i] == sub_fragments[i]:
				master_leftovers = master_fragments[i+1:]
				sub_leftovers = sub_fragments[i+1:]
		except IndexError:
			break

	pathstr = ''
	for f in master_leftovers[:-1]:
		pathstr += '..' + s
	pathstr += s.join(sub_leftovers)
	return pathstr

def reverse_root_subfile(masterfile, subfile):
	"""
	does inverse operation to root_subfile. E.g. 
	E.g. if masterfile is ./../foo/bar.xml and subfile is ../foo2/subfoo.xml,
	returned path ./../foo2/subfoo.xml
	Usually this function is used to convert saved paths into engine relative paths
	NOTE: masterfile is expected to be *file*, not directory. subfile can be either
	"""
	s = '/'

	masterfile = norm_path(os.path.abspath(masterfile)).split(s)[:-1]
	subfile = norm_path(os.path.abspath( s.join(masterfile) + s + subfile ))
	masterfile = norm_path(os.getcwd()) + s + 'foo.bar' # cheat a little to satisfy root_subfile
	return root_subfile(masterfile, subfile)

def norm_path(path):
	"""
	Makes the path use '/' delimited separators. FIFE always uses these delimiters, but some os-related
	routines will default to os.path.sep.
	"""
	if os.path.sep == '/':
		return path

	return '/'.join(path.split(os.path.sep))


def frange(limit1, limit2 = None, increment = 1.):
	"""Range function that accepts floats (and integers).
	If only one limit is specified, assumes 0 as lower limit.

	Usage:
	frange(-2, 2, 0.1)
	frange(10)
	frange(10, increment = 0.5)

	The returned value is an iterator.  Use list(frange) for a list.

	source: U{http://code.activestate.com/recipes/
	66472-frange-a-range-function-with-float-increments/}
 
	@type	limit1:	float
	@param	limit1:	lower range limit
	@type	limit2:	float
	@param	limit2:	upper range limit
	@type	increment:	float
	@param	increment:	length of each step
	@rtype	generator
	@return	iterable over (limit2 - limit1) / increment steps
	"""

	if limit2 is None:
		limit2, limit1 = float(limit1), 0.
	else:
		limit1 = float(limit1)

	count = int(math.ceil(old_div((limit2 - limit1),increment)))
	return (limit1 + n*increment for n in range(count))
	



# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2019 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

### Widget/Container Base Classes ###

"""
Widget wrappers.

Please look at the documentation of L{Widget} for details.
"""
from __future__ import absolute_import

from .widget import Widget

from .spacer import Spacer
from .adjustingcontainer import AdjustingContainer
from .containers import Container, ABox, VBox, HBox, CBox, Window
from .flowcontainer import FlowContainer
from .label import Label
from .icon import Icon
from .animationicon import AnimationIcon
from .buttons import Button, ToggleButton, ImageButton
from .checkbox import CheckBox
from .radiobutton import RadioButton
from .tabbedarea import TabbedArea, Tab
from .textfield import TextField
from .textbox import TextBox
from .listbox import ListBox
from .dropdown import DropDown
from .scrollarea import ScrollArea
from .slider import Slider
from .percentagebar import PercentageBar
from .passwordfield import PasswordField
from .iconprogressbar import IconProgressBar
from .imageprogressbar import ImageProgressBar
from .resizablewindow import ResizableWindow
from .dockarea import DockArea
from .panel import Panel
from .bargraph import BarGraph
from .curvegraph import CurveGraph
from .linegraph import LineGraph
from .piegraph import PieGraph
from .pointgraph import PointGraph


# Global Widget Class registry

WIDGETS = {
	# Containers
	"Container" : Container,
	"Window" : Window,
	"ABox" : ABox,
	"VBox" : VBox,
	"HBox" : HBox,
	"CBox" : CBox,
	"ScrollArea" :ScrollArea,
	"AdjustingContainer" : AdjustingContainer,
	"FlowContainer" : FlowContainer,
	"TabbedArea" : TabbedArea,
	"Tab" : Tab,
	"ResizableWindow" : ResizableWindow,
	"DockArea" : DockArea,
	"Panel" : Panel,

	# Simple Widgets
	"Spacer" : Spacer,
	"Icon" : Icon,
	"AnimationIcon" : AnimationIcon,
	"Label" : Label,
	"PercentageBar" : PercentageBar,
	"IconProgressBar" : IconProgressBar,
	"ImageProgressBar" : ImageProgressBar,
	"BarGraph" : BarGraph,
	"CurveGraph" : CurveGraph,
	"LineGraph" : LineGraph,
	"PieGraph" : PieGraph,
	"PointGraph" : PointGraph,

	# Button Widgets
	"Button" : Button,
	"CheckBox" : CheckBox,
	"RadioButton" : RadioButton,
	"ImageButton" : ImageButton,
	"ToggleButton" : ToggleButton,

	#Complexer Widgets / Text io
	"TextField" : TextField,
	"TextBox" : TextBox,
	"PasswordField" : PasswordField,
	"ListBox" : ListBox,
	"DropDown" : DropDown,
	"Slider" : Slider
}

def registerWidget(cls):
	"""
	Register a new Widget class for pychan.
	
	If the widget class name is already used you will overwrite the existing
	definition.  There is no warning to tell you if this has happened.  Be sure
	to understand what you are doing here!
	"""
	global WIDGETS
	name = cls.__name__

	WIDGETS[name] = cls



# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2019 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

"""
The basic application and main loop.

See the L{ApplicationBase} documentation.
"""

from fife import fife
from fife.extensions import fifelog
from fife.extensions.basicapplication import ApplicationBase
from fife.extensions import pychan
from fife.extensions.fife_settings import Setting

class ExitEventListener(fife.IKeyListener):
	"""
	Default, rudimentary event listener.

	Will cause the application to quit on pressing ESC.
	"""
	def __init__(self, app):
		self.app = app
		self.engine = app.engine
		eventmanager = self.engine.getEventManager()
		#eventmanager.setNonConsumableKeys([fife.Key.ESCAPE])
		fife.IKeyListener.__init__(self)
		eventmanager.addKeyListener(self)
		self.quitRequested = False

	def keyPressed(self, evt):
		keyval = evt.getKey().getValue()
		if keyval == fife.Key.ESCAPE:
			self.app.quit()
		elif keyval == fife.Key.F10:
			if pychan.manager.hook.guimanager.isConsoleEnabled():
				pychan.manager.hook.guimanager.getConsole().toggleShowHide()
				evt.consume()

	def keyReleased(self, evt):
		pass

class PychanApplicationBase(ApplicationBase):
	"""
	PychanApplicationBase is an extendable class that provides a basic environment for a FIFE-based client.
	This class should be extended if you 've build fife with fifechan support, in order to use pychan.
	"""
	def __init__(self, setting=None):
		super(PychanApplicationBase, self).__init__(setting)
		
		pychan.init(self.engine, debug = self._finalSetting['PychanDebug'])
		pychan.setupModalExecution(self.mainLoop,self.breakFromMainLoop)
		
	def createListener(self):
		self._listener = ExitEventListener(self)
		return self._listener



# -*- coding: utf-8 -*-
# ####################################################################
#  Copyright (C) 2005-2019 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

""" submodule for xml map parsing """
from __future__ import print_function

from builtins import str
from builtins import object
from fife import fife

from fife.extensions.serializers import ET
from fife.extensions.serializers import SerializerError, InvalidFormat
from fife.extensions.serializers import NameClash, NotFound, WrongFileType
from fife.extensions.serializers.xmlanimation import loadXMLAnimation

class XMLObjectSaver(object):
	""" The B{XMLObjectSaver} serializes a fife.Object instance by saving
	it back to its XML file

	@note:
		- this code does NOT allow the creation of a new xml file
		- this code does NOT touch atlas or animation definitions
		- this code does NOT allow saving to non-well-formed xml files
		- this code DOES save blocking & static flag, as well as
		  image offsets

	@type	engine:	fife
	@ivar	engine:	pointer to initialized fife engine instance
	@type	img_manager:	fife.ImageManager
	@ivar	img_manager:	pointer to fife image manager
	@type	compat:	bool
	@ivar	compat:	flag to either use outdated xml definitions or new approach
	@type	debug:	bool
	@ivar	debug:	flag to activate/deactivate debug output
	@type	vfs:	fife.VFS
	@ivar	vfs:	pointer to fife vfs
	@type	change:	bool
	@ivar	change:	flag if object data differs from file data
	"""
	PROCESSING_INSTRUCTION = '<?fife type="object"?>'
	def __init__(self, engine, debug=False, compat=True):
		"""

		@type	engine:	fife
		@param	engine:	intialized fife engine
		"""
		self.compat = compat
		self.debug = debug
		self.engine = engine
		self.img_manager = engine.getImageManager()
		self.vfs = self.engine.getVFS()
		self.change = False

	def save(self, object):
		""" saves the data of a fife.Object to its xml file

		@type	object:	fife.Object
		@param	object:	the object which should be saved
		@rtype	bool
		@return	flag whether the saving was successful or not
		"""
		self.change = False
		result = False

		file = object.getFilename()
		if not file:
			raise SerializerError("Object cannot be saved, no file found %s" % object)
			return result

		if not self.vfs.exists(file):
			raise NotFound("File not within vfs: %s" % file)
			return result

		file_handle = self.vfs.open(file)
		file_handle.thisown = 1
		tree = ET.parse(file_handle)
		root = tree.getroot()

		object_id = object.getId()
		blocking = object.isBlocking()
		static = object.isStatic()

		cost_id = object.getCostId()
		cost = object.getCost()
		cellstack_pos = object.getCellStackPosition()

		if self.debug:
			print("XML tree dump: (pre-save)")
			ET.dump(root)
			print("Object data: ")
			print("\tid", object_id)
			print("\tblocking", blocking)
			print("\tstatic", static)
			print("\tcost id", cost_id)
			print("\tcost", cost)

		# check for compat mode
		if root.tag != 'assets':
			self.compat = True

		# in compat mode tree is <object>
		if self.compat:
			objects = [root,]
		# new XML structure has tree root <assets> which groups multiple objects
		else:
			objects = root.findall("object")

		for obj in objects:
			_id = obj.get("id")
			if _id != object_id:
				if self.debug:
					print("...ommitting object %s " % _id)
				continue

			if int(obj.attrib['blocking']) != int(blocking):
				self.change = True
			if int(obj.attrib['static']) != int(static):
				self.change = True
			if str(obj.attrib['cost_id']) != str(cost_id):
				self.change = True
			if float(obj.attrib['cost']) != float(cost):
				self.change = True
			if int(obj.attrib['cellstack_position']) != int(cellstack_pos):
				self.change = True

			obj.attrib['blocking'] = str(int(blocking))
			obj.attrib['static'] = str(int(static))
			obj.attrib['cost_id'] = str(cost_id)
			obj.attrib['cost'] = str(cost)
			obj.attrib['cellstack_position'] = str(cellstack_pos)

			if self.debug and self.change:
				print("\tSet new data in xml tree: ")
				print("\t\tblocking: ", obj.attrib['blocking'])
				print("\t\tstatic: ", obj.attrib['static'])

			images = obj.findall("image")
			actions = obj.findall("action")

			if self.debug:
				print("\tAttempting to save image data: ")
				print("\t...found these image elements: ")
				print("\t", images)
				print("object dump: ")
				print(ET.dump(obj))

			self.save_images(images, object)
			self.save_actions(actions, object)

		if not self.change:
			return result

		xmlcontent = ET.tostring(root)

		if self.debug:
			print("XML tree dump: (post-manipulation)")
			ET.dump(root)

		# save xml data beneath the <?fife type="object"?> definition into the object file
		file = open(file, 'w')
		file.write(XMLObjectSaver.PROCESSING_INSTRUCTION+'\n')
		file.write(xmlcontent + "\n")
		file.close()
		result = True
		return result

	def save_actions(self, actions, object):
		""" save action definitions

		@type	actions:	list
		@param	actions:	list of <action> elements
		@type	object:	fife.Object
		@param	object:	the object which should be saved
		"""
		for element in actions:
			# new xml format uses this, we only save the new format
			if 'animation_id' not in element.attrib:
				break

			animation_id = element.attrib['animation_id']
			self.save_animation(animation_id, object)

	def save_animation(self, animation_id, object):
		""" save animation definitions for the given id

		@type	animation_id:	str
		@param	animation_id:	id of the animation data structure
		@type	object:	fife.Object
		@param	object:	the object which should be saved
		"""
		pass

	def save_images(self, images, object):
		"""	save image definitions

		@type	images:	list
		@param	images:	list of <image> elements
		@type	object:	fife.Object
		@param	object:	the object which should be saved
		"""
		visual = object.get2dGfxVisual()
		angles = visual.getStaticImageAngles()
		if self.debug:
			print("\t\tobject angles: ", angles)

		for element in images:
			angle = int(element.get("direction"))
			if angle not in angles: continue

			index = visual.getStaticImageIndexByAngle(angle)
			image = self.img_manager.get(index)
			x_offset = image.getXShift()
			y_offset = image.getYShift()

			if 'x_offset' not in element.attrib or int(element.attrib['x_offset']) != x_offset:
				self.change = True
			if 'y_offset' not in element.attrib or int(element.attrib['y_offset']) != y_offset:
				self.change = True

			element.attrib['x_offset'] = str(x_offset)
			element.attrib['y_offset'] = str(y_offset)

			if self.debug and self.change:
				print("\tSet new data in xml tree: (<image>) ")
				print("\t\tx offset: ", element.attrib['x_offset'])
				print("\t\ty offset: ", element.attrib['y_offset'])

class XMLObjectLoader(object):
	"""

	"""
	def __init__(self, engine):
		"""

		"""
		self.engine = engine
		self.imgMgr = engine.getImageManager()
		self.anim_pool = None
		self.model = engine.getModel()
		self.vfs = engine.getVFS()
		self.source = None
		self.filename = ''

	def loadResource(self, location):
		"""

		"""
		self.source = location
		self.filename = self.source
		self.node = None
		self.file = None
		if hasattr(location, 'node'):
			self.node = location.node
		else:
			isobjectfile = True
			f = self.vfs.open(self.filename)
			f.thisown = 1

			obj_identifier = '<?fife type="object"?>'
			try:
				s = f.readString(len(obj_identifier))
			except fife.IndexOverflow:
				isobjectfile = False

			if isobjectfile and not s.startswith(obj_identifier):
				isobjectfile = False

			if not isobjectfile:
				return

				# this will never be hit currently, if this is put before the return it can provide useful debugging
				# but animation.xml files will raise this exception because apparently they come through here first
				raise WrongFileType('Tried to open non-object file %s with XMLObjectLoader.' % self.filename)

		self.do_load_resource(f)

	def do_load_resource(self, file):
		"""

		"""
		if file:
			tree = ET.parse(file)
			self.node = tree.getroot()
		self.parse_object(self.node)

	def parse_object(self, object):
		"""

		"""
		if self.node.tag != 'object':
			raise InvalidFormat('Expected <object> tag, but found <%s>.' % self.node.tag)

		_id = object.get('id')
		if not _id:
			raise InvalidFormat('<object> declared without an id attribute.')
		_id = str(_id)

		nspace = object.get('namespace')
		if not nspace:
			raise InvalidFormat('<object> %s declared without a namespace attribute.' % str(_id))
		nspace = str(nspace)

		obj = None
		parent = object.get('parent', None)
		if parent:
			query = self.metamodel.getObjects('id', str(parent))
			if len(query) == 0:
				raise NotFound('No objects found with identifier %s.' % str(parent))
			elif len(query) > 1:
				raise NameClash('%d objects found with identifier %s.' % (len(query), str(parent)))
			parent = query[0]

		# check if model already has this object
		if not bool(self.model.getObject(_id, nspace)):
			obj = self.model.createObject(_id, nspace, parent)
		else:
			print(NameClash('Tried to create already existing object \n\t...ignoring: %s, %s' % (_id, nspace)))
			return

		obj.setFilename(self.source)
		fife.ObjectVisual.create(obj)
		obj.setBlocking(bool( int(object.get('blocking', False)) ))
		obj.setStatic(bool( int(object.get('static', False)) ))

		pather = object.get('pather', 'RoutePather')
		obj.setPather( self.model.getPather(pather) )

		self.parse_images(object, obj)
		self.parse_actions(object, obj)

	def parse_images(self, objelt, object):
		"""

		"""
		for image in objelt.findall('image'):
			source = image.get('source')
			if not source:
				raise InvalidFormat('<image> declared without a source attribute.')

			# paths are relative to this resource's path
			path = self.filename.split('/')
			path.pop()
			path.append(str(source))

			img = self.imgMgr.create('/'.join(path))
			img.setXShift(int( image.get('x_offset', 0) ))
			img.setYShift(int( image.get('y_offset', 0) ))

			object.get2dGfxVisual().addStaticImage(int( image.get('direction', 0) ), img.getHandle())

	def parse_actions(self, objelt, object):
		"""

		"""
		for action in objelt.findall('action'):
			id = action.get('id')
			if not id:
				raise InvalidFormat('<action> declared without an id attribute.')

			act_obj = object.createAction(str(id))
			fife.ActionVisual.create(act_obj)
			self.parse_animations(action, act_obj)

	def parse_animations(self, actelt, action):
		"""

		"""
		pass
		for anim in actelt.findall('animation'):
			source = anim.get('source')
			if not source:
				raise InvalidFormat('Animation declared with no source location.')

			# animation paths are relative to this resource's path
			path = self.filename.split('/')
			path.pop()
			path.append(str(source))

			animation = loadXMLAnimation(self.engine, '/'.join(path))
			action.get2dGfxVisual().addAnimation(int( anim.get('direction', 0) ), animation)
			action.setDuration(animation.getDuration())



# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2019 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

"""
Simple error checking attributes.

This module defines a set of Attribute classes
which you can use to define possible values
an attribute of an object accepts.

Usage::
  class SomeObject:
      nameAttr, posAttr = [ Attr("name"), PointAttr("pos") ]

  obj =  SomeObject()
  obj.nameAttr.set(obj,"newName")
  obj.posAttr.set(obj,"89,89")

This is most useful for error checking parsing and defining
accepted attributes in classes and is used by pychan internally.

"""
from __future__ import absolute_import

from builtins import map
from builtins import str
from builtins import object
from .exceptions import ParserError

class Attr(object):
	"""
	A simple text attribute.
	"""
	def __init__(self,name):
		self.name = name

	def set(self,obj,value):
		"""
		Parses the given value with the L{parse} method
		and sets it on the given instance with C{setattr}.
		"""
		value = self.parse(value)
		setattr(obj,self.name,value)

	def parse(self,value):
		"""
		Parses a value and checks for errors.
		Override with specialiced behaviour.
		"""
		return str(value)

class UnicodeAttr(Attr):
	def parse(self,value):
		"""
		Parses a value and checks for errors.
		Override with specialiced behaviour.
		"""
		return str(value)


class PointAttr(Attr):
	def parse(self,value):
		try:
			x,y = tuple(map(int,str(value).split(',')))
			return x,y
		except:
			raise ParserError(str(self.name)+" expected a comma separated list of two integers.")

class ColorAttr(Attr):
	def parse(self,value):
		a = 255
		try:
			try:
				r,g,b,a = tuple(map(int,str(value).split(',')))
				for c in (r,g,b,a):
					if not 0 <= c < 256: raise ParserError(str(self.name)+" expected a color (Failed: 0 <= %d <= 255)" %c)
			except:
				r,g,b = tuple(map(int,str(value).split(',')))
				for c in (r,g,b):
					if not 0 <= c < 256: raise ParserError(str(self.name)+" expected a color (Failed: 0 <= %d <= 255)" %c)
		except:
			raise ParserError(str(self.name)+" expected a color.")

		return r,g,b,a

class IntAttr(Attr):
	def parse(self,value):
		try:
			return int(value)
		except:
			raise ParserError(str(self.name)+" expected a single integer.")

class BoolAttr(Attr):
	def parse(self,value):
		try:
			# Allow clients to be compatible with old FIFE versions
			if value in ["0", "False", "false"]:
				value = False
			elif value in ["1", "True", "true"]:
				value = True
			if value not in (True,False):
				raise ParserError(str(self.name)+" expected False or True.")
			return value
		except:
			raise ParserError(str(self.name)+" expected False or True.")

class FloatAttr(Attr):
	def parse(self, value):
		try:
			return float(value)
		except:
			raise ParserError(str(self.name)+" expected a float.")


class ListAttr(Attr):
	def parse(self, value):
		try:
			result = list(map(str,str(value).split(',')))
			return result
		except:
			raise ParserError(str(self.name)+" expected a list with strings.")

class UnicodeListAttr(Attr):
	def parse(self, value):
		try:
			result = list(map(unicode,str(value).split(',')))
			return result
		except:
			raise ParserError(str(self.name)+" expected a list with unicode strings.")

class IntListAttr(Attr):
	def parse(self, value):
		try:
			result = list(map(int,str(value).split(',')))
			return result
		except:
			raise ParserError(str(self.name)+" expected a list with ints.")

class BoolListAttr(Attr):
	def parse(self, value):
		try:
			result = list(map(bool,str(value).split(',')))
			return result
		except:
			raise ParserError(str(self.name)+" expected a list with bools.")

class FloatListAttr(Attr):
	def parse(self, value):
		try:
			result = list(map(float,str(value).split(',')))
			return result
		except:
			raise ParserError(str(self.name)+" expected a list with floats.")


ATTRIBUTE_TYPE = { "Str" : Attr,
				   "Unicode" : UnicodeAttr,
				   "Point" : PointAttr,
				   "Color" : ColorAttr,
				   "Int" : IntAttr,
				   "Bool" : BoolAttr,
				   "Float" : FloatAttr
				 }	
class MixedListAttr(Attr):
	def parse(self, value):
		try:
			result = []
			# we use ';' to split, because of point and color attributes
			# example: "Int:5; Color:255,0,255"
			tmp_result = map(str,str(value).split(';'))
			for r in tmp_result:
				type = str(r).split(':')[0].strip()
				val = str(r).split(':')[1].strip()

				m = ATTRIBUTE_TYPE.get(type)
				if m is None:
					raise ParserError("Unknown attribute type "+type+".")
				method = getattr(m(self), 'parse')
				result.append(method(val))
			return result
		except:
			raise ParserError(str(self.name)+" expected a mixed list.")



# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2019 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

""" a filebrowser implementation for pychan """
from __future__ import print_function

from builtins import str
from builtins import object
import sys
import os

from fife.fife import Engine
import fife.extensions.pychan.widgets as widgets
import fife.extensions.pychan as pychan


def u2s(string):
	return string.encode(sys.getfilesystemencoding())


class FileBrowser(object):
	""" The B{FileBrowser} displays directory and file listings.

		B{The fileSelected} parameter is a callback invoked when a file selection has been made; its
		signature must be fileSelected(path,filename).

		B{If selectdir} is set, fileSelected's filename parameter should be optional.

		B{The savefile} option provides a box for supplying a new filename that doesn't exist yet.

		B{The selectdir} option allows directories to be selected as well as files.
	"""

	def __init__(self, *args, **kwargs):
		"""Temporary constructor for deprecation"""
		#TODO: Remove the engine argument
		display_deprecation = False
		if isinstance(args[0], Engine):
			display_deprecation = True
			args = args[1:]
		if "engine" in kwargs:
			display_deprecation = True
			del kwargs["engine"]
		if display_deprecation:
			print("DEPRECATION WARNING:")
			print("The filebrowser no longer needs an engine instance")
			print("It will be completely removed in the future")

		self.__real_init__(*args, **kwargs)

	def __real_init__(self, fileSelected, savefile=False, selectdir=False, extensions=('xml',), guixmlpath="gui/filebrowser.xml"):
		"""
		@type	fileSelected:	function
		@param	fileSelected:	callback invoked on file selection
		@type	savefile:	bool
		@param	savefile:	flag to provide a gui for an usergiven filename
		@type	selectdir:	bool
		@param	selectdir:	flag to fire fileSelected without filename
		@type	extensions:	tuple
		@param	extensions:	list of extensions the filebrowser should show (defaults to xml)
		@type	guixmlpath:	string
		@param	guixmlpath:	path to the xml guifile defaults to (gui/filebrowser.xml)
		"""
		self.fileSelected = fileSelected

		self._widget = None
		self.savefile = savefile
		self.selectdir = selectdir

		self.guixmlpath = guixmlpath

		self.extensions = extensions
		self.path = './..'
		self.dir_list = []
		self.file_list = []

	def showBrowser(self):
		"""	create and / or show the gui """
		if self._widget:
			self.setDirectory(self.path)
			self._widget.show()
			return

		self._widget = pychan.loadXML(self.guixmlpath)
		self._widget.mapEvents({
			'dirList'	   : self._selectDir,
			'selectButton'  : self._selectFile,
			'closeButton'   : self._widget.hide
		})
		if self.savefile:
			self._file_entry = widgets.TextField(name='saveField', text=u'')
			self._widget.findChild(name="fileColumn").addChild(self._file_entry)

		self.setDirectory(self.path)
		self._widget.show()

	def setDirectory(self, path):
		""" sets the current directory according to path """
		path_copy = self.path
		self.path = path
		if not self._widget: return

		def decodeList(list):
			fs_encoding = sys.getfilesystemencoding()
			if fs_encoding is None: fs_encoding = "ascii"

			newList = []
			for i in list:
				try: newList.append(str(i, fs_encoding))
				except:
					newList.append(str(i, fs_encoding, 'replace'))
					print("WARNING: Could not decode item:", i)
			return newList

		dir_list_copy = list(self.dir_list)
		file_list_copy = list(self.file_list)

		self.dir_list = []
		self.file_list = []

		contents = os.listdir(path)
		for content in contents:
			if os.path.isdir(os.path.join(path, content)):
				self.dir_list.append(content)
			elif os.path.isfile(os.path.join(path, content)):
				extension = os.path.splitext(content)[1][1:]
				if extension in self.extensions:
					self.file_list.append(content)

		try:
			self.dir_list = sorted(decodeList(self.dir_list), key=lambda s: s.lower())
			self.dir_list.insert(0, "..")
			self.file_list = sorted(decodeList(self.file_list), key=lambda s: s.lower())
		except:
			self.path = path_copy
			self.dir_list = list(dir_list_copy)
			self.file_list = list(file_list_copy)
			print("WARNING: Tried to browse to directory that is not accessible!")

		self._widget.distributeInitialData({
			'dirList'  : self.dir_list,
			'fileList' : self.file_list
		})

		self._widget.adaptLayout()

	def _selectDir(self):
		""" callback for directory ListBox """
		selection = self._widget.collectData('dirList')
		if selection >= 0 and selection < len(self.dir_list):
			new_dir = u2s(self.dir_list[selection])
			lst = self.path.split('/')
			if new_dir == '..' and lst[-1] != '..' and lst[-1] != '.':
				lst.pop()
			else:
				lst.append(new_dir)

			path = '/'.join(lst)
			self.setDirectory(path)

	def _selectFile(self):
		""" callback for selectButton, hides filebrowser """
		self._widget.hide()
		selection = self._widget.collectData('fileList')

		if self.savefile:
			if self._widget.collectData('saveField'):
				self.fileSelected(self.path, u2s(self._widget.collectData('saveField')))
				return

		if selection >= 0 and selection < len(self.file_list):
			self.fileSelected(self.path, u2s(self.file_list[selection]))
			return

		if self.selectdir:
			self.fileSelected(self.path)
			return

		print('FileBrowser: error, no selection.')



# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2019 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

from __future__ import absolute_import
from fife import fifechan

from fife.extensions.pychan.attrs import BoolAttr

from .basictextwidget import BasicTextWidget


class Label(BasicTextWidget):
	"""
	A basic label - displaying a string.

	Also allows text wrapping and onMouse hover callbacks.

	New Attributes
	==============

	 - wrap_text: Boolean: Enable/Disable automatic text wrapping. Disabled by default.
	 Currently to actually see text wrapping you have to explicitly set a max_size with
	 the desired width of the text, as the layout engine is not capable of deriving
	 the maximum width from a parent container.

	"""

	ATTRIBUTES = BasicTextWidget.ATTRIBUTES + [ BoolAttr('wrap_text') ]

	DEFAULT_WRAP_TEXT = False

	def __init__(self, 
				 parent = None,
				 name = None,
				 size = None,
				 min_size = None,
				 max_size = None,
				 fixed_size = None,
				 margins = None,
				 padding = None,
				 helptext = None,
				 position = None,
				 style = None,
				 hexpand = None,
				 vexpand = None,
				 font = None,
				 base_color = None,
				 background_color = None,
				 foreground_color = None,
				 selection_color = None,
				 border_color = None,
				 outline_color = None,
				 border_size = None,
				 outline_size = None,
				 position_technique = None,
				 is_focusable = None,
				 comment = None,
				 text = None,
				 wrap_text = None):
				 
		self.real_widget = fifechan.Label("")
		self.wrap_text = self.DEFAULT_WRAP_TEXT
		super(Label,self).__init__(parent=parent, 
								   name=name, 
								   size=size, 
								   min_size=min_size, 
								   max_size=max_size,
								   fixed_size=fixed_size,
								   margins=margins,
								   padding=padding,
								   helptext=helptext, 
								   position=position,
								   style=style, 
								   hexpand=hexpand, 
								   vexpand=vexpand,
								   font=font,
								   base_color=base_color,
								   background_color=background_color,
								   foreground_color=foreground_color,
								   selection_color=selection_color,
								   border_color=border_color,
								   outline_color=outline_color,
								   border_size=border_size,
								   outline_size=outline_size,
								   position_technique=position_technique,
								   is_focusable=is_focusable,
								   comment=comment,
								   text=text)
								   
		if wrap_text is not None:
			self.wrap_text = wrap_text
		
	def clone(self, prefix):
		lblClone = Label(None,
						 self._createNameWithPrefix(prefix),
						 self.size,
						 self.min_size,
						 self.max_size,
						 self.fixed_size,
						 self.margins,
						 self.padding,
						 self.helptext,
						 self.position,
						 self.style,
						 self.hexpand,
						 self.vexpand,
						 self.font,
						 self.base_color,
						 self.background_color,
						 self.foreground_color,
						 self.selection_color,
						 self.border_color,
						 self.outline_color,
						 self.border_size,
						 self.outline_size,
						 self.position_technique,
						 self.is_focusable,
						 self.comment,
						 self.text,
						 self.wrap_text)
		
		return lblClone;

	def _setTextWrapping(self,wrapping): self.real_widget.setTextWrapping(wrapping)
	def _getTextWrapping(self): self.real_widget.isTextWrapping()
	wrap_text = property(_getTextWrapping,_setTextWrapping)



# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2013 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

from __future__ import absolute_import
from fife import fife
from fife import fifechan

from fife.extensions.pychan.attrs import IntAttr, UnicodeAttr

from .containers import Container

class FlowContainer(Container):
	"""
	An implementation of a flow container that can contain other widgets.
	The widgets can be sorted vertical per row or horizontal per column.
	If the space in the container is too small to put all the components in one row or column,
	it uses multiple rows or columns.

	"""

	ATTRIBUTES = Container.ATTRIBUTES + [ IntAttr('alignment'), ]

	DEFAULT_LAYOUT = 'Horizontal'
	DEFAULT_ALIGNMENT = 4 # Center

	def __init__(self, 
				 parent = None,
				 name = None,
				 size = None,
				 min_size = None,
				 max_size = None,
				 fixed_size = None,
				 margins = None,
				 padding = None,
				 helptext = None,
				 position = None,
				 style = None,
				 hexpand = None,
				 vexpand = None,
				 font = None,
				 base_color = None,
				 background_color = None,
				 foreground_color = None,
				 selection_color = None,
				 border_color = None,
				 outline_color = None,
				 border_size = None,
				 outline_size = None,
				 position_technique = None,
				 is_focusable = None,
				 comment = None,
				 background_image = None,
				 opaque = None,
				 layout = None,
				 spacing = None,
				 uniform_size = None,
				 _real_widget = None,
				 alignment = None):

		if _real_widget is None: _real_widget = fifechan.FlowContainer()
		
		super(FlowContainer,self).__init__(parent=parent,
										   name=name,
										   size=size,
										   min_size=min_size,
										   max_size=max_size,
										   fixed_size=fixed_size,
										   margins=margins,
										   padding=padding,
										   helptext=helptext,
										   position=position,
										   style=style,
										   hexpand=hexpand,
										   vexpand=vexpand,
										   font=font,
										   base_color=base_color,
										   background_color=background_color,
										   foreground_color=foreground_color,
										   selection_color=selection_color,
										   border_color=border_color,
										   outline_color=outline_color,
										   border_size=border_size,
										   outline_size=outline_size,
										   position_technique=position_technique,
										   is_focusable=is_focusable,
										   comment=comment,
										   background_image=background_image,
										   opaque=opaque,
										   layout=layout,
										   spacing=spacing,
										   uniform_size=uniform_size,
										   _real_widget=_real_widget)

		if alignment is not None: self.alignment = alignment
		else: self.alignment = self.DEFAULT_ALIGNMENT

				
	def clone(self, prefix):
		containerClone = FlowContainer(None, 
						self._createNameWithPrefix(prefix),
						self.size,
						self.min_size,
						self.max_size,
						self.fixed_size,
						self.margins,
						self.padding,
						self.helptext,
						self.position,
						self.style,
						self.hexpand,
						self.vexpand,
						self.font,
						self.base_color,
						self.background_color,
						self.foreground_color,
						self.selection_color,
						self.border_color,
						self.outline_color,
						self.border_size,
						self.outline_size,
						self.position_technique,
						self.is_focusable,
						self.comment,
						self.background_image,
						self.opaque,
						self.layout,
						self.spacing,
						self.uniform_size,
						None,
						self.alignment)
			
		containerClone.addChildren(self._cloneChildren(prefix))	
		return containerClone

	def _setAlignment(self, alignment): self.real_widget.setAlignment(alignment)
	def _getAlignment(self): self.real_widget.getAlignment()
	alignment = property(_getAlignment, _setAlignment)

	def adjustContent(self):
		self.real_widget.adjustContent()



# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2019 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

from __future__ import print_function
from __future__ import absolute_import
from builtins import str
from builtins import object
import weakref

from fife import fife
from fife import fifechan

from fife.extensions.pychan import events
from fife.extensions.pychan.attrs import (Attr, UnicodeAttr, PointAttr,
                                          ColorAttr, BoolAttr, IntAttr, IntListAttr)
from fife.extensions.pychan.exceptions import StopTreeWalking
from fife.extensions.pychan.properties import ColorProperty

from .common import get_manager
from .layout import isLayouted


class Widget(object):
	"""
	This is the common widget base class, which provides most of the wrapping
	functionality.

	Attributes
	==========

	Widgets are manipulated (mostly) through attributes - and these can all be set by XML attributes.
	Derived widgets will have other attributes. Please see their B{New Attributes} sections. The types of the
	attributes are pretty straightforward, but note that Position and Color attribute types will also accept
	C{fife.Point} and C{fife.Color} values.

	  - name: String: The identification of the widget, most useful if it is unique within a given widget hiarachy.
	  This is used to find widgets by L{mapEvents},L{distributeInitialData},L{distributeData} and L{collectData}.
	  - position: Position: The position relative to the parent widget - or on screen, if this is the root widget.
	  - size: Position: The real size of the widget (including border and margins). Usually you do not need to set this.
	  A notable exception is the L{ScrollArea}.
	  - min_size: Position: The minimal size this widget is allowed to have. This is enforced through the accessor methods
	  of the actual size attribute.
	  - max_size: Position: The maximal size this widget is allowed to have. This is enforced through the accessor methods
	  of the actual size attribute.
	  - margins: Integer list: The margin clears an area around an element (outside the border) and is completely transparent.
	  - padding: Integer list: The padding clears an area around the content (inside the border) of the element.
	  - base_color: Color
	  - background_color: Color
	  - foreground_color: Color
	  - selection_color: Color
	  - border_color: Color
	  - outline_color: Color
	  - font: String: This should identify a font that was loaded via L{loadFonts} before.
	  - helptext: Unicode: Text which can be used for e.g. tooltips.
	  - comment: Unicode: Additional text stored by the widget.  Not used by PyChan directly. Can be used by the client for additional info about the widget.
	  - border_size: Integer: The size of the border in pixels.
	  - position_technique: This can be either "automatic" or "explicit" - only L{Window} has this set to "automatic" which
	  results in new windows being centered on screen (for now).
	  If it is set to "explicit" the position attribute will not be touched.
	  - vexpand: Bool: True. Proportion to expand this widget vertically.
	  - hexpand: Bool: True. Proportion to expand this widget horizontally.

	Convenience Attributes
	======================

	These attributes are convenience/shorthand versions of above mentioned attributes and assignment will reflect
	the associated attributes values. E.g. the following is equivalent::
	   # Set X position, leave Y alone
	   widget.x = 10
	   # Same here
	   posi = widget.position
	   widget.position = (10, posi[1])

	Here they are.

	   - x: Integer: The horizontal part of the position attribute.
	   - y: Integer: The vertical part of the position attribute.
	   - width: Integer: The horizontal part of the size attribute.
	   - height: Integer: The vertical part of the size attribute.

	"""

	ATTRIBUTES = [ Attr('name'), 
				   PointAttr('position'),
				   PointAttr('min_size'), 
				   PointAttr('size'), 
				   PointAttr('max_size'),
				   PointAttr('fixed_size'),
				   IntListAttr('margins'),
				   IntListAttr('padding'),
				   ColorAttr('base_color'),
				   ColorAttr('background_color'),
				   ColorAttr('foreground_color'),
				   ColorAttr('selection_color'),
				   ColorAttr('border_color'),
				   ColorAttr('outline_color'),
				   Attr('style'), 
				   Attr('font'),
				   IntAttr('border_size'),
				   IntAttr('outline_size'),
				   Attr('position_technique'),
				   BoolAttr('vexpand'),
				   BoolAttr('hexpand'), 
				   UnicodeAttr('helptext'),
				   BoolAttr('is_focusable'), 
				   UnicodeAttr('comment')
		]

	DEFAULT_NAME = '__unnamed__'
	DEFAULT_HEXPAND = False
	DEFAULT_VEXPAND = False
	DEFAULT_MAX_SIZE = 500000, 500000
	DEFAULT_SIZE = -1, -1
	DEFAULT_MIN_SIZE = 0, 0
	DEFAULT_MARGINS = 0, 0
	DEFAULT_PADDING = 0
	DEFAULT_HELPTEXT = u""
	DEFAULT_POSITION = 0, 0
	DEFAULT_FONT = "default"
	DEFAULT_BORDER_SIZE = 0
	DEFAULT_OUTLINE_SIZE = 0
	DEFAULT_POSITION_TECHNIQUE = "explicit"
	DEFAULT_COMMENT = u""

	HIDE_SHOW_ERROR = """\
		You can only show/hide the top widget of a hierachy.
		Use 'addChild' or 'removeChild' to add/remove labels for example.
		"""

	def __init__(self,
				 parent = None, 
				 name = None,
				 size = None,
				 min_size = None, 
				 max_size = None,
				 fixed_size = None,
				 margins = None,
				 padding = None,
				 helptext = None, 
				 position = None, 
				 style = None, 
				 hexpand = None,
				 vexpand = None,
				 font = None,
				 base_color = None,
				 background_color = None,
				 foreground_color = None,
				 selection_color = None,
				 border_color = None,
				 outline_color = None,
				 border_size = None,
				 outline_size = None,
				 position_technique = None,
				 is_focusable = None,
				 comment = None):
		
		# Make sure the real_widget has been created
		assert( hasattr(self,'real_widget') )

		self.event_mapper = events.EventMapper(self)

		# Flag to indicate if the Widget is added to the Manager
		self._added = False
		# Flag to indicate if the Widget is added to
		# the top Widget list of the Manager
		self._top_added = False
		# Only needed for tabs
		self.tab = None
		
		# Data distribution & retrieval settings
		self.accepts_data = False
		self.accepts_initial_data = False
		
		#set all defaults
		if get_manager().compat_layout:
			self.hexpand, self.vexpand = 0,0
		else:
			self.hexpand = self.DEFAULT_HEXPAND
			self.vexpand = self.DEFAULT_VEXPAND 		

		self.name = self.DEFAULT_NAME
		self.has_name = False
		self.position = self.DEFAULT_POSITION
		self.position_technique = self.DEFAULT_POSITION_TECHNIQUE
		self.font = self.DEFAULT_FONT
		self.min_size = self.DEFAULT_MIN_SIZE
		self.max_size = self.DEFAULT_MAX_SIZE
		self.size = self.DEFAULT_SIZE
		self.margins = self.DEFAULT_MARGINS
		self.padding = self.DEFAULT_PADDING
		self.border_size = self.DEFAULT_BORDER_SIZE
		self.outline_size = self.DEFAULT_OUTLINE_SIZE
		self.helptext = self.DEFAULT_HELPTEXT
		self.comment = self.DEFAULT_COMMENT
		self._usedPrefixes = []
	
		# Parent attribute makes sure we only have one parent,
		# that tests self.__parent - so make sure we have the attr here.
		self.__parent = None
		self.parent = parent

		# Inherit and apply style
		if style is None and parent:
			style = parent.style
		self.style = style or "default"
		
		# override everything style has set
		if vexpand is not None:	self.vexpand = vexpand
		if hexpand is not None: self.hexpand = hexpand
		if name is not None: 
			self.name = name
			self.has_name = True

		if position is not None: self.position = position
		if position_technique is not None: self.position_technique = position_technique
		if font is not None: self.font = font
		
		# only set this if it's provided
		if is_focusable is not None: self.is_focusable = is_focusable
		
		if min_size is not None: self.min_size = min_size
		if max_size is not None: self.max_size = max_size
		if size is not None: self.size = size
		if fixed_size is not None: self.fixed_size = fixed_size
		if margins is not None: self.margins = margins
		if padding is not None: self.padding = padding
		if border_size is not None: self.border_size = border_size
		if outline_size is not None: self.outline_size = outline_size
		
		if helptext is not None: self.helptext = helptext
		if comment is not None: self.comment = comment

		# these are set in the default style
		if base_color is not None: self.base_color = base_color
		if background_color is not None: self.background_color = background_color
		if foreground_color is not None: self.foreground_color = foreground_color
		if selection_color is not None: self.selection_color = selection_color
		if border_color is not None: self.border_color = border_color
		if outline_color is not None: self.outline_color = outline_color
		# add this widget to the manager
		get_manager().addWidget(self)
	
	def clone(self, prefix):
		"""
		Clones this widget.
		
		Concrete widgets should implement this one, if not, an exception should
		be raised.
		
		Prefix is used to create the name of the cloned widget.
		"""
		raise RuntimeError("No implementation of clone method for %s" % self.__class__)
			      

	def execute(self, bind, focus=None):
		"""
		Execute a dialog synchronously.

		As argument a dictionary mapping widget names to return values
		is expected. Events from these widgets will cause this function
		to return with the associated return value.

		This function will not return until such an event occurs.
		The widget will be shown before execution and hidden afterwards.
		You can only execute root widgets.

		@param focus: name of child widget which should have focus. Defaults to main widget.

		Note: This feature is not tested well, and the API will probably
		change. Otherwise have fun::
		  # Okay this a very condensed example :-)
		  return pychan.loadXML("contents/gui/dialog.xml").execute({ 'okButton' : True, 'closeButton' : False })

		"""
		if not get_manager().can_execute:
			raise RuntimeError("Synchronous execution is not set up!")
		if self.__parent:
			raise RuntimeError("You can only 'execute' root widgets, not %s!" % str(self))

		for name,returnValue in list(bind.items()):
			def _quitThisDialog(returnValue = returnValue ):
				get_manager().breakFromMainLoop( returnValue )
				self.hide()
			self.findChild(name=name).capture( _quitThisDialog , group_name = "__execute__" )
		self.show()
		if focus and self.findChild(name=focus):
			self.findChild(name=focus).is_focusable = True
			self.findChild(name=focus).requestFocus()
		else:
			self.is_focusable = True
			self.requestFocus()
		return get_manager().mainLoop()

	def requestFocus(self):
		"""
		Requests focus.  
		
		The widget must be focusable in order for this to work.  See 
		the is_focusable property.
		
		"""
		if self.isVisible():
			self.real_widget.requestFocus()

	def isModalFocusable(self):
		"""
		Checks if a widget is modal focusable.
		True if no other widget has modal focus, false otherwise.
		"""
		return self.real_widget.isModalFocusable()

	def isModalFocused(self):
		"""
		Checks if the widget or it's parent has modal focus.
		"""
		return self.real_widget.isModalFocused()
	
	def requestModalFocus(self):
		"""
		Requests modal focus.
		When a widget has modal focus, only that widget and it's children may recieve input.
		
		The widget must be modal focusable in order for this to work. Therefore,
		no other widget should has modal focus.
		
		"""
		if self.isVisible():
			if self.isModalFocusable():
				self.real_widget.requestModalFocus()

	def releaseModalFocus(self):
		"""
		Releases modal focus.
		"""
		if self.isModalFocused():
			self.real_widget.releaseModalFocus()

	def isModalMouseInputFocusable(self):
		"""
		Checks if a widget is modal mouse input focusable.
		True if no other widget has modal mouse input focus, false otherwise.
		"""
		return self.real_widget.isModalMouseInputFocusable()

	def isModalMouseInputFocused(self):
		"""
		Checks if the widget or it's parent has modal mouse input focus.
		"""
		return self.real_widget.isModalMouseInputFocused()
	
	def requestModalMouseInputFocus(self):
		"""
		Requests modal mouse input focus.
		When a widget has modal input focus that widget will be the only widget
		receiving input even if the input occurs outside of the widget and
		no matter what the input is.
		
		The widget must be modal mouse input focusable in order for this to work.
		Therefore, no other widget should has modal input focus.
		
		"""
		if self.isVisible():
			if self.isModalMouseInputFocusable():
				self.real_widget.requestModalMouseInputFocus()

	def releaseModalMouseInputFocus(self):
		"""
		Releases modal mouse input focus.
		"""
		if self.isModalMouseInputFocused():
			self.real_widget.releaseModalMouseInputFocus()
			
	def match(self,**kwargs):
		"""
		Matches the widget against a list of key-value pairs.
		Only if all keys are attributes and their value is the same it returns True.
		"""
		for k,v in list(kwargs.items()):
			if v != getattr(self,k,None):
				return False
		return True

	def capture(self, callback, event_name="action", group_name="default"):
		"""
		Add a callback to be executed when the widget event occurs on this widget.

		The callback must be either a callable or None.
		The old event handler (if any) will be overridden by the callback.
		If None is given, the event will be disabled. You can query L{isCaptured}
		wether this widgets events are currently captured.

		It might be useful to check out L{tools.callbackWithArguments}.

		@param callback: Event callback - may accept keyword arguments event and widget.
		@param event_name: The event to capture - may be one of L{events.EVENTS} and defaults to "action"
		@param group_name: Event group.

		Event groups are used to have different B{channels} which don't interfere with each other.
		For derived widgets that need to capture events it's advised to use the group_name 'widget'.
		The 'default' group is used by default, and should be reserved for the application programmers.
		"""
		self.event_mapper.capture( event_name, callback, group_name )

	def isCaptured(self):
		"""
		Check whether this widgets events are captured
		(a callback is installed) or not.
		"""
		return bool(self.event_mapper.getCapturedEvents())

	def show(self):
		"""
		Show the widget and all contained widgets.
		"""
		# add this widget to the manager
		if not self._added:
			get_manager().addWidget(self)
		if self.parent is None and not self._top_added:
			get_manager().addTopWidget(self)
		# add childs of this widget to the manager
		def _show(shown_widget):
			get_manager().addWidget(shown_widget)
		self.deepApply(_show)
		
		if self.isVisible() and self.isSetVisible():
			self.beforeShow()
			self.adaptLayout()
			if self.parent is None:
				get_manager().placeWidget(self, self.position_technique)
			return

		self.beforeShow()
		# Show real widget to distribute a widgetShown event.
		self.real_widget.setVisible(True)
		self.adaptLayout()
		if self.parent is None:
			get_manager().placeWidget(self, self.position_technique)

	def hide(self, free=False):
		"""
		Hide the widget and all contained widgets.
		"""
		# remove this widget from the manager
		if self._added:
			get_manager().removeWidget(self)
		if self.parent is None and self._top_added:
			get_manager().removeTopWidget(self)
		# remove childs of this widget from the manager
		def _hide(hidden_widget):
			get_manager().removeWidget(hidden_widget)
		self.deepApply(_hide)
		
		if not self.isVisible() and not self.isSetVisible():
			self.adaptLayout()
			self.afterHide()
			return

		# Hide real widget to distribute a widgetHidden event.
		self.real_widget.setVisible(False)

		if free:
			if self.parent:
				self.parent.removeChild(self)
			self.removeAllChildren()

		self.adaptLayout()
		self.afterHide()

	def isVisible(self):
		"""
		Check whether the widget is currently shown,
		either directly or as part of a container widget.
		"""
		return self.real_widget.isVisible()

	def isSetVisible(self):
		"""
		Check the real widget visible flag.
		It checks not if the widget is currently shown!
		This is needed e.g. if the parent is already hidden
		but we want to hide the child too.
		"""
		return self.real_widget.isSetVisible()
	
	def adaptLayout(self,recurse=True):
		"""
		Execute the Layout engine. Automatically called by L{show}.
		In case you want to relayout a visible widget.
		This function will automatically perform the layout adaption
		from the top-most layouted widget.

		To make this clear consider this arrangement::
			VBox 1
				- Container
			VBox 2
				- HBox
					- Label

		If you call adaptLayout on the Label the layout from the VBox 2
		will get recalculated, while the VBox 1 stays untouched.

		@param recurse: Pass False here to force the layout to start from
		this widget.
		"""
		self.real_widget.adaptLayout(recurse)

	def beforeShow(self):
		"""
		This method is called just before the widget is shown.
		You can override this in derived widgets to add finalization
		behaviour.
		
		NOTE:
			- if your widget is a container, you have to call
			  _resetTiling(), as you will loose this call by using
			  your override method
		"""

	def afterHide(self):
		"""
		This method is called just before the widget is hidden.
		You can override this in derived widgets to add finalization
		behaviour.
		"""

	def findChildren(self,**kwargs):
		"""
		Find all contained child widgets by attribute values.

		Usage::
		  closeButtons = root_widget.findChildren(name='close')
		  buttons = root_widget.findChildren(__class__=pychan.widgets.Button)
		"""

		children = []
		def _childCollector(widget):
			if widget.match(**kwargs):
				children.append(widget)
		self.deepApply(_childCollector)
		return children

	def getNamedChildren(self, include_unnamed = False):
		"""
		Create a dictionary of child widgets with the keys being
		their name. This will contain only Widgets which have
		a name different from "__unnamed__" (which is the default).
		
		@param include_unnamed: Defaults to false. If this is true unnamed widgets are added, too.
		
		The values are lists of widgets, so not only unique names
		are handled correctly.

		Usage::
			children = widget.getNamedChildren()
			for widget in children.get("info",[])
				print widget.name , " == info"
		"""
		children = {}
		if include_unnamed:
			def _childCollector(widget):
				children.setdefault(widget._name,[]).append(widget)
		else:
			def _childCollector(widget):
				if widget.has_name:
					children.setdefault(widget._name,[]).append(widget)
		self.deepApply(_childCollector)
		return children

	def findChild(self,**kwargs):
		""" Find the first contained child widgets by attribute values.

		Usage::
		  closeButton = root_widget.findChild(name='close')
		"""
		if list(kwargs.keys()) == ["name"]:
			return self.findChildByName(kwargs["name"])

		children = self.findChildren(**kwargs)
		if children:
			return children[0]
		return None

	def findChildByName(self,name):
		"""
		Find first contained child widget by its name.

		Note that this is the fast version of findChild(name="...")
		and that you don't have to call this explicitly, it is used
		if possible.
		"""
		result = []
		def _childCollector(widget):
			if widget._name == name:
				result.append(widget)
				raise StopTreeWalking
		try:
			self.deepApply(_childCollector)
		except StopTreeWalking:
			return result[0]
		return None

	def addChild(self,widget):
		"""
		This function adds a widget as child widget and is only implemented
		in container widgets.

		You'll need to call L{adaptLayout} if the container is already shown,
		to adapt the layout to the new widget. This doesn't happen
		automatically.
		"""
		raise RuntimeError("Trying to add a widget to %s, which doesn't allow this." % repr(self))

	def insertChild(self, widget, position):
		"""
		This function inserts a widget a given index in the child list.
		
		See L{addChild} and L{insertChildBefore} 
		"""
		raise RuntimeError("Trying to insert a widget to %s, which doesn't allow this." % repr(self))
		
	def insertChildBefore(self, widget, before):
		"""
		Inserts a child widget before a given widget. If the widget isn't found,
		the widget is appended to the children list.
		
		See L{addChild} and L{insertChild} 
		"""
		raise RuntimeError("Trying to insert a widget to %s, which doesn't allow this." % repr(self))

	def addChildren(self,*widgets):
		"""
		Add multiple widgets as children.
		Only implemented for container widgets. See also L{addChild}

		Usage::
			container.addChildren( widget1, widget2, ... )
			# or you can use this on a list
			container.addChildren( [widget1,widget2,...] )
		"""
		if len(widgets) == 1 and not isinstance(widgets[0],Widget):
			widgets = widgets[0]
		for widget in widgets:
			self.addChild(widget)

	def removeChild(self,widget):
		"""
		This function removes a direct child widget and is only implemented
		in container widgets.

		You'll need to call L{adaptLayout} if the container is already shown,
		to adapt the layout to the removed widget. This doesn't happen
		automatically.
		"""
		raise RuntimeError("Trying to remove a widget from %s, which is not a container widget." % repr(self))

	def removeChildren(self,*widgets):
		"""
		Remove a list of direct child widgets.
		All widgets have to be direct child widgets.
		To 'clear' a container take a look at L{removeAllChildren}.
		See also L{removeChild}.

		Usage::
			container.removeChildren( widget1, widget2, ... )
			# or you can use this on a list
			container.removeChildren( [widget1,widget2,...] )
		"""
		if len(widgets) == 1 and not isinstance(widgets[0],Widget):
			widgets = widgets[0]
		for widget in widgets:
			self.removeChild(widget)

	def removeAllChildren(self):
		"""
		This function will remove all direct child widgets.
		This will work even for non-container widgets.
		"""
		children = self.findChildren(parent=self)
		for widget in children:
			self.removeChild(widget)

	def mapEvents(self,eventMap,ignoreMissing = False):
		"""
		Convenience function to map widget events to functions
		in a batch.

		Subsequent calls of mapEvents will merge events with different
		widget names and override the previously set callback.
		You can also pass C{None} instead of a callback, which will
		disable the event completely.

		@param eventMap: A dictionary with widget/event names as keys and callbacks as values.
		@param ignoreMissing: Normally this method raises a RuntimeError, when a widget
		can not be found - this behaviour can be overriden by passing True here.

		The keys in the dictionary are parsed as C{"widgetName/eventName"} with the slash
		separating the two. If no slash is found the eventName is assumed to be "action".

		Additionally you can supply a group name or channel C{"widgetName/eventName/groupName"}.
		Event handlers from one group are not overridden by handlers from another group.
		The default group name is C{"default"}.

		Example::
			guiElement.mapEvents({
				"button" : guiElement.hide,
				"button/mouseEntered" : toggleButtonColorGreen,
				"button/mouseExited" :  toggleButtonColorBlue,
			})

		"""
		children = self.getNamedChildren(include_unnamed=True)
		for descr,func in list(eventMap.items()):
			name, event_name, group_name = events.splitEventDescriptor(descr)
			#print name, event_name, group_name
			widgets = children.get(name,[])
			if widgets:
				for widget in widgets:
					widget.capture( func, event_name = event_name, group_name = group_name )
			elif not ignoreMissing:
				raise RuntimeError("No widget with the name: %s" % name)

	def setInitialData(self,data):
		"""
		Set the initial data on a widget, what this means depends on the Widget.
		In case the widget does not accept initial data, a L{RuntimeError} is thrown.
		"""
		if not self.accepts_initial_data:
			raise RuntimeError("Trying to set data on a widget that does not accept initial data. Widget: %s Data: %s " % (repr(self),repr(data)))
		self._realSetInitialData(data)

	def setData(self,data):
		"""
		Set the user-mutable data on a widget, what this means depends on the Widget.
		In case the widget does not accept data, a L{RuntimeError} is thrown.
		This is inverse to L{getData}.
		"""
		if not self.accepts_data:
			raise RuntimeError("Trying to set data on a widget that does not accept data.")
		self._realSetData(data)

	def getData(self):
		"""
		Get the user-mutable data of a widget, what this means depends on the Widget.
		In case the widget does not have user mutable data, a L{RuntimeError} is thrown.
		This is inverse to L{setData}.
		"""
		if not self.accepts_data:
			raise RuntimeError("Trying to retrieve data from a widget that does not accept data.")
		return self._realGetData()

	def distributeInitialData(self,initialDataMap):
		"""
		Distribute B{initial} (not mutable by the user) data from a dictionary over the widgets in the hierachy
		using the keys as names and the values as the data (which is set via L{setInitialData}).
		If more than one widget matches - the data is set on ALL matching widgets.
		By default a missing widget is just ignored.

		Use it like this::
		  guiElement.distributeInitialData({
		       'myTextField' : 'Hello World!',
		       'myListBox' : ["1","2","3"]
		  })

		"""
		children = self.getNamedChildren(include_unnamed=True)
		for name,data in list(initialDataMap.items()):
			widgetList = children.get(name,[])
			for widget in widgetList:
				widget.setInitialData(data)

	def distributeData(self,dataMap):
		"""
		Distribute data from a dictionary over the widgets in the hierachy
		using the keys as names and the values as the data (which is set via L{setData}).
		This will only accept unique matches.

		Use it like this::
		  guiElement.distributeData({
		       'myTextField' : 'Hello World!',
		       'myListBox' : ["1","2","3"]
		  })

		"""
		children = self.getNamedChildren(include_unnamed=True)
		for name,data in list(dataMap.items()):
			widgetList = children.get(name,[])
			if len(widgetList) != 1:
				if get_manager().debug:
					self.listNamedWidgets()
				raise RuntimeError("DistributeData can only handle widgets with unique names.")
			widgetList[0].setData(data)

	def collectDataAsDict(self,widgetNames):
		"""
		Collect data from a widget hierachy by names into a dictionary.
		This can only handle UNIQUE widget names (in the hierachy)
		and will raise a RuntimeError if the number of matching widgets
		is not equal to one.

		Usage::
		  data = guiElement.collectDataAsDict(['myTextField','myListBox'])
		  print "You entered:",data['myTextField']," and selected ",data['myListBox']

		"""
		children = self.getNamedChildren(include_unnamed=True)
		dataMap = {}
		for name in widgetNames:
			widgetList = children.get(name,[])
			if len(widgetList) != 1:
				if get_manager().debug:
					self.listNamedWidgets()
				raise RuntimeError("CollectData can only handle widgets with unique names.")

			dataMap[name] = widgetList[0].getData()
		return dataMap

	def collectData(self,*widgetNames):
		"""
		Collect data from a widget hierachy by names.
		This can only handle UNIQUE widget names (in the hierachy)
		and will raise a RuntimeError if the number of matching widgets
		is not equal to one.

		This function takes an arbitrary number of widget names and
		returns a list of the collected data in the same order.

		In case only one argument is given, it will return just the
		data, with out putting it into a list.

		Usage::
		  # Multiple element extraction:
		  text, selected = guiElement.collectData('myTextField','myListBox')
		  print "You entered:",text," and selected item nr",selected
		  # Single elements are handled gracefully, too:
		  test = guiElement.collectData('testElement')

		"""
		children = self.getNamedChildren(include_unnamed=True)
		dataList = []
		for name in widgetNames:
			widgetList = children.get(name,[])
			if len(widgetList) != 1:
				if get_manager().debug:
					self.listNamedWidgets()
				raise RuntimeError("CollectData can only handle widgets with unique names.")
			dataList.append( widgetList[0].getData() )
		if len(dataList) == 1:
			return dataList[0]
		return dataList

	def listNamedWidgets(self):
		"""
		This function will print a list of all currently named child-widgets
		to the standard output. This is useful for debugging purposes.
		"""
		def _printNamedWidget(widget):
			if widget.name != Widget.DEFAULT_NAME:
				print(widget.name.ljust(20),repr(widget).ljust(50),repr(widget.__parent))
		print("Named child widgets of ",repr(self))
		print("name".ljust(20),"widget".ljust(50),"parent")
		self.deepApply(_printNamedWidget)

	def stylize(self,style,**kwargs):
		"""
		Recursively apply a style to all widgets.
		"""
		def _restyle(widget):
			get_manager().stylize(widget,style,**kwargs)
		self.deepApply(_restyle)

	def resizeToContent(self,recurse = True):
		"""
		Try to shrink the widget, so that it fits closely around its content.
		Do not call directly.
		"""
		self.real_widget.resizeToContent(recurse)

	def expandContent(self,recurse = True):
		"""
		Try to expand any spacer in the widget within the current size.
		Do not call directly.
		"""
		self.real_widget.expandContent(recurse)

	def _recursiveResizeToContent(self):
		"""
		Recursively call L{resizeToContent}. Uses L{deepApply}.
		Do not call directly.
		"""
		self.real_widget.resizeToContent(True)

	def _recursiveExpandContent(self):
		"""
		Recursively call L{expandContent}. Uses L{deepApply}.
		Do not call directly.
		"""
		self.real_widget.expandContent(True)

	def deepApply(self,visitorFunc, leaves_first = True, shown_only = False):
		"""
		Recursively apply a callable to all contained widgets and then the widget itself.
		"""
		visitorFunc(self)
		
	def getAbsolutePos(self):
		"""
		Get absolute position on screen
		"""
		absX = self.x
		absY = self.y
		parent = self.parent
		while parent is not None:
			absX += parent.x
			absY += parent.y
			parent = parent.parent
		return (absX, absY)

	def sizeChanged(self):
		pass

	def __str__(self):
		return "%s(name='%s')" % (self.__class__.__name__,self.name)

	def __repr__(self):
		return "<%s(name='%s') at %x>" % (self.__class__.__name__,self.name,id(self))

	def _setSize(self,size):
		if isinstance(size,fife.Point):
			self.width, self.height = size.x, size.y
		else:
			self.width, self.height = size

	def _getSize(self):
		return self.width, self.height

	def _setPosition(self,size):
		if isinstance(size,fife.Point):
			self.x, self.y = size.x, size.y
		else:
			self.x, self.y = size

	def _getPosition(self):
		return self.x, self.y

	def _setX(self,x):self.real_widget.setX(x)
	def _getX(self): return self.real_widget.getX()
	def _setY(self,y): self.real_widget.setY(y)
	def _getY(self): return self.real_widget.getY()

	def _setWidth(self,w):
		old_width = self.width
		self.real_widget.setWidth(w)
		w = self.real_widget.getWidth()
		if w != old_width:
			self.sizeChanged()

	def _getWidth(self): return self.real_widget.getWidth()
	def _setHeight(self,h):
		old_height = self.height
		self.real_widget.setHeight(h)
		h = self.real_widget.getHeight()
		if h != old_height:
			self.sizeChanged()

	def _getHeight(self): return self.real_widget.getHeight()

	def _getMinWidth(self): return self.min_size[0]
	def _getMaxWidth(self): return self.max_size[0]
	def _getMinHeight(self): return self.min_size[1]
	def _getMaxHeight(self): return self.max_size[1]
	def _setMinWidth(self,w):
		self.min_size = w, self.min_size[1]
	def _setMaxWidth(self,w):
		self.max_size = w, self.max_size[1]
	def _setMinHeight(self,h):
		self.min_size = self.min_size[0],h
	def _setMaxHeight(self,h):
		self.max_size = self.max_size[0],h

	def _setMinSize(self, size):
		self.real_widget.setMinSize(fifechan.Size(size[0], size[1]))
	def _getMinSize(self):
		size = self.real_widget.getMinSize()
		return (size.getWidth(), size.getHeight())
	def _setMaxSize(self, size):
		self.real_widget.setMaxSize(fifechan.Size(size[0], size[1]))
	def _getMaxSize(self):
		size = self.real_widget.getMaxSize()
		return (size.getWidth(), size.getHeight())
	def _setFixedSize(self, size):
		self.real_widget.setFixedSize(fifechan.Size(size[0], size[1]))
	def _getFixedSize(self):
		size = self.real_widget.getFixedSize()
		return (size.getWidth(), size.getHeight())
	def isFixedSize(self):
		return self.real_widget.isFixedSize()
	
	def _setFont(self, font):
		self._font = font
		self.real_font = get_manager().getFont(font)
		self.real_widget.setFont(self.real_font)
	def _getFont(self):
		return self._font

	def _getOutlineSize(self):
		return self.real_widget.getOutlineSize()
	def _setOutlineSize(self, size):
		self.real_widget.setOutlineSize(size)
		
	def _getBorderSize(self):
		return self.real_widget.getBorderSize()
	def _setBorderSize(self, size):
		self.real_widget.setBorderSize(size)

	def _setMargins(self, margin):
		# Shorthand property
		if isinstance(margin, tuple) or isinstance(margin, list):
			if len(margin) == 4:
				# 0=top, 1=right, 2=bottom, 3=left
				self.real_widget.setMarginTop(margin[0])
				self.real_widget.setMarginRight(margin[1])
				self.real_widget.setMarginBottom(margin[2])
				self.real_widget.setMarginLeft(margin[3])
			elif len(margin) == 3:
				# 0=top, 1=right, 2=bottom, 1=left
				self.real_widget.setMarginTop(margin[0])
				self.real_widget.setMarginRight(margin[1])
				self.real_widget.setMarginBottom(margin[2])
				self.real_widget.setMarginLeft(margin[1])
			elif len(margin) == 2:
				# 0=top, 1=right, 0=bottom, 1=left
				self.real_widget.setMarginTop(margin[0])
				self.real_widget.setMarginRight(margin[1])
				self.real_widget.setMarginBottom(margin[0])
				self.real_widget.setMarginLeft(margin[1])
			elif len(margin) == 1:
				# 0=top, 0=right, 0=bottom, 0=left
				self.real_widget.setMargin(margin[0])
		else:
			self.real_widget.setMargin(margin)
	def _getMargins(self):
		return (self.real_widget.getMarginTop(), self.real_widget.getMarginRight(),
				self.real_widget.getMarginBottom(), self.real_widget.getMarginLeft())
	
	def _setPadding(self, padding):
		# Shorthand property
		if isinstance(padding, tuple) or isinstance(padding, list):
			if len(padding) == 4:
				# 0=top, 1=right, 2=bottom, 3=left
				self.real_widget.setPaddingTop(padding[0])
				self.real_widget.setPaddingRight(padding[1])
				self.real_widget.setPaddingBottom(padding[2])
				self.real_widget.setPaddingLeft(padding[3])
			elif len(padding) == 3:
				# 0=top, 1=right, 2=bottom, 1=left
				self.real_widget.setPaddingTop(padding[0])
				self.real_widget.setPaddingRight(padding[1])
				self.real_widget.setPaddingBottom(padding[2])
				self.real_widget.setPaddingLeft(padding[1])
			elif len(padding) == 2:
				# 0=top, 1=right, 0=bottom, 1=left
				self.real_widget.setPaddingTop(padding[0])
				self.real_widget.setPaddingRight(padding[1])
				self.real_widget.setPaddingBottom(padding[0])
				self.real_widget.setPaddingLeft(padding[1])
			elif len(padding) == 1:
				# 0=top, 0=right, 0=bottom, 0=left
				self.real_widget.setPadding(padding[0])
		else:
			self.real_widget.setPadding(padding)
	def _getPadding(self):
		return (self.real_widget.getPaddingTop(), self.real_widget.getPaddingRight(),
				self.real_widget.getPaddingBottom(), self.real_widget.getPaddingLeft())
	
	base_color = ColorProperty("BaseColor")
	background_color = ColorProperty("BackgroundColor")
	foreground_color = ColorProperty("ForegroundColor")
	selection_color = ColorProperty("SelectionColor")
	outline_color = ColorProperty("OutlineColor")
	border_color = ColorProperty("BorderColor")
	
	def _getStyle(self): return self._style
	def _setStyle(self,style):
		self._style = style
		get_manager().stylize(self,style)
	style = property(_getStyle,_setStyle)

	def _getParent(self):
		if self.__parent is not None:
			return self.__parent()
		return None
	def _setParent(self,parent):
		if parent and not issubclass(type(parent), Widget):
			raise RuntimeError("Parent must be subclass of the Widget type.")

		if self.__parent is not None and self.__parent() is not parent:
			if self.__parent() is not None and parent is not None:
				print("Widget containment fumble:", self, self.__parent, parent)
				self.__parent().removeChild(self)
		if parent is not None:
			self.__parent = weakref.ref(parent)
		else:
			self.__parent = None
	parent = property(_getParent,_setParent)

	def _setName(self,name):
		self._name = name
		if name != Widget.DEFAULT_NAME:
			self.has_name = True
	def _getName(self):
		# __str__ relies on self.name
		return getattr(self,'_name','__no_name_yet__')
	name = property(_getName,_setName)

	def _setFocusable(self, b): self.real_widget.setFocusable(b)
	def _isFocusable(self):
		return self.real_widget.isFocusable()

	def _setHExpand(self, expand):
		self.real_widget.setHorizontalExpand(expand)
	def _isHExpand(self):
		return self.real_widget.isHorizontalExpand()
	def _setVExpand(self, expand):
		self.real_widget.setVerticalExpand(expand)
	def _isVExpand(self):
		return self.real_widget.isVerticalExpand()
	
	def _createNameWithPrefix(self, prefix):
		
		if not isinstance(prefix, str):
			raise RuntimeError("Widget names should be prefixed with a string")
		
		if prefix in self._usedPrefixes:
			raise RuntimeError("Widget %s already cloned with prefix %s" % (self.name, prefix))
		
		if len(prefix) == 0:
			raise RuntimeError("New widget name cannot be created with an empty prefix")
		
		self._usedPrefixes.append(prefix)
		
		return prefix + self.name

	x = property(_getX,_setX)
	y = property(_getY,_setY)
	width = property(_getWidth,_setWidth)
	height = property(_getHeight,_setHeight)
	min_width = property(_getMinWidth,_setMinWidth)
	min_height = property(_getMinHeight,_setMinHeight)
	max_width = property(_getMaxWidth,_setMaxWidth)
	max_height = property(_getMaxHeight,_setMaxHeight)
	size = property(_getSize,_setSize)
	min_size = property(_getMinSize, _setMinSize)
	max_size = property(_getMaxSize, _setMaxSize)
	fixed_size = property(_getFixedSize, _setFixedSize)
	position = property(_getPosition,_setPosition)
	font = property(_getFont,_setFont)
	outline_size = property(_getOutlineSize,_setOutlineSize)
	border_size = property(_getBorderSize,_setBorderSize)
	is_focusable = property(_isFocusable,_setFocusable)
	margins = property(_getMargins, _setMargins)
	padding = property(_getPadding, _setPadding)
	is_focusable = property(_isFocusable,_setFocusable)
	hexpand = property(_isHExpand,_setHExpand)
	vexpand = property(_isVExpand,_setVExpand)



# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2013 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

from __future__ import absolute_import
from fife import fife
from fife import fifechan

from fife.extensions.pychan.attrs import IntAttr, IntListAttr

from .containers import Container

class AdjustingContainer(Container):
	"""
	This is the adjusting container class. It provides space in which child widgets can
	be position via the position attribute.

	New Attributes
	==============

	  - colums - Integer: The number of columns to divide the widgets into.
	  - alignments: The alignment per column.
	"""

	ATTRIBUTES = Container.ATTRIBUTES + [ IntAttr('columns'),
										  IntListAttr('alignments'),
										]
	DEFAULT_NUMBER_COLUMNS = 3
	
	def __init__(self, 
				 parent = None,
				 name = None,
				 size = None,
				 min_size = None,
				 max_size = None,
				 fixed_size = None,
				 margins = None,
				 padding = None,
				 helptext = None,
				 position = None,
				 style = None,
				 hexpand = None,
				 vexpand = None,
				 font = None,
				 base_color = None,
				 background_color = None,
				 foreground_color = None,
				 selection_color = None,
				 border_color = None,
				 outline_color = None,
				 border_size = None,
				 outline_size = None,
				 position_technique = None,
				 is_focusable = None,
				 comment = None,
				 background_image = None,
				 opaque = None,
				 layout = None,
				 spacing = None,
				 uniform_size = None,
				 _real_widget = None,
				 columns = None,
				 alignments = None):

		if _real_widget is None: _real_widget = fifechan.AdjustingContainer()
		
		super(AdjustingContainer,self).__init__(parent=parent,
												name=name,
												size=size,
												min_size=min_size,
												max_size=max_size,
												fixed_size=fixed_size,
												margins=margins,
												padding=padding,
												helptext=helptext,
												position=position,
												style=style,
												hexpand=hexpand,
												vexpand=vexpand,
												font=font,
												base_color=base_color,
												background_color=background_color,
												foreground_color=foreground_color,
												selection_color=selection_color,
												border_color=border_color,
												outline_color=outline_color,
												border_size=border_size,
												outline_size=outline_size,
												position_technique=position_technique,
												is_focusable=is_focusable,
												comment=comment,
												background_image=background_image,
												opaque=opaque,
												layout=layout,
												spacing=spacing,
												uniform_size=uniform_size,
												_real_widget=_real_widget)

		if columns is not None:
			self.columns = columns
		else:
			self.columns = self.DEFAULT_NUMBER_COLUMNS

		if alignments is not None:
			self.alignments = alignments
				
	def clone(self, prefix):
		containerClone = AdjustingContainer(None, 
						self._createNameWithPrefix(prefix),
						self.size,
						self.min_size,
						self.max_size,
						self.fixed_size,
						self.margins,
						self.padding,
						self.helptext,
						self.position,
						self.style,
						self.hexpand,
						self.vexpand,
						self.font,
						self.base_color,
						self.background_color,
						self.foreground_color,
						self.selection_color,
						self.border_color,
						self.outline_color,
						self.border_size,
						self.outline_size,
						self.position_technique,
						self.is_focusable,
						self.comment,
						self.background_image,
						self.opaque,
						self.layout,
						self.spacing,
						self.uniform_size,
						None,
						self.columns,
						self.alignments)
			
		containerClone.addChildren(self._cloneChildren(prefix))	
		return containerClone
		
	def _setNumberColumns(self, number):
		self.real_widget.setNumberOfColumns(number)
	def _getNumberColumns(self):
		self.real_widget.getNumberOfColumns()
	columns = property(_getNumberColumns, _setNumberColumns)
	
	def _setColumnAlignment(self, column, alignment):
		self.real_widget.setColumnAlignment(column, alignment)
	def _getColumnAlignment(self, column):
		self.real_widget.getColumnAlignment(column)
	column_alignment = property(_getColumnAlignment, _setColumnAlignment)

	def _setColumnAlignments(self, alignments):
		i = 0
		if alignments is not None:
			for a in alignments:
				self.real_widget.setColumnAlignment(i, a)
				i += 1
		else:
			cols = self.columns
			while i < cols:
				self.real_widget.setColumnAlignment(i, 0)
				i += 1
	def _getColumnAlignments(self):
		alignments = []
		cols = self.columns
		i = 0
		while i < cols:
			alignments.append(self.column_alignment(i))
			i += 1
		return alignments
	alignments = property(_getColumnAlignments, _setColumnAlignments)




# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2013 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

from __future__ import absolute_import
from fife import fife
from fife import fifechan

from fife.extensions.pychan.attrs import BoolAttr, ColorAttr, UnicodeAttr
from fife.extensions.pychan.properties import ColorProperty
from .common import get_manager

from .resizablewindow import ResizableWindow

class DockArea(ResizableWindow):
	"""
	This is a window class that can be resizable and it's possible to dock Panels.

	New Attributes
	==============

	  - active: If true, the DockArea is active and Panels can be docked to it.
	  - side: Allows to specify the side for expanding.
	  - highlight_color: Color that is used if a Panel enters the DockArea.
	"""

	ATTRIBUTES = ResizableWindow.ATTRIBUTES + [ BoolAttr('active'),
												UnicodeAttr('side'),
												ColorAttr('highlight_color'),
												]

	DEFAULT_MARGINS = 0
	DEFAULT_PADDING = 0
	DEFAULT_SPACING = 0
	
	DEFAULT_ACTIVE = True
	DEFAULT_HIGHLIGHT_COLOR = (200, 0, 0)

	def __init__(self, 
				 parent = None, 
				 name = None,
				 size = None,
				 min_size = None, 
				 max_size = None,
				 fixed_size = None,
				 margins = None,
				 padding = None,
				 helptext = None,
				 position = None, 
				 style = None, 
				 hexpand = None,
				 vexpand = None,
				 font = None,
				 base_color = None,
				 background_color = None,
				 foreground_color = None,
				 selection_color = None,
				 border_color = None,
				 outline_color = None,
				 border_size = None,
				 outline_size = None,
				 position_technique = None,
				 is_focusable = None,
				 comment = None,
				 background_image = None,
				 opaque = None,
				 layout = None,
				 spacing = None,
				 uniform_size = None,
				 _real_widget = None,
				 title = None,
				 titlebar_height = None,
				 movable = None,
				 resizable = None,
				 top_resizable = None,
				 right_resizable = None,
				 bottom_resizable = None,
				 left_resizable = None,
				 shove = None,
				 cursors = None,
				 active = None,
				 side = None,
				 highlight_color = None):
		
		if _real_widget is None: _real_widget = fifechan.DockArea()
			
		super(DockArea,self).__init__(parent=parent,
									  name=name,
									  size=size,
									  min_size=min_size,
									  max_size=max_size,
									  fixed_size=fixed_size,
									  margins=margins,
									  padding=padding,
									  helptext=helptext,
									  position=position,
									  style=style,
									  hexpand=hexpand,
									  vexpand=vexpand,
									  font=font,
									  base_color=base_color,
									  background_color=background_color,
									  foreground_color=foreground_color,
									  selection_color=selection_color,
									  border_color=border_color,
									  outline_color=outline_color,
									  border_size=border_size,
									  outline_size=outline_size,
									  position_technique=position_technique,
									  is_focusable=is_focusable,
									  comment=comment,
									  background_image=background_image,
									  opaque=opaque,
									  layout=layout,
									  spacing=spacing,
									  uniform_size=uniform_size,
									  _real_widget=_real_widget,
									  title=title,
									  titlebar_height=titlebar_height,
									  movable=movable,
									  resizable=resizable,
									  top_resizable=top_resizable,
									  right_resizable=right_resizable,
									  bottom_resizable=bottom_resizable,
									  left_resizable=left_resizable,
									  shove=shove,
									  cursors=cursors)

		if active is not None: self.active = active
		else: self.active = self.DEFAULT_ACTIVE
		
		if side is not None: self.side = side

		if highlight_color is not None: self.highlight_color = highlight_color
		else: self.highlight_color = self.DEFAULT_HIGHLIGHT_COLOR

	def clone(self, prefix):
		dockAreaClone = DockArea(None, 
					self._createNameWithPrefix(prefix),
					self.size,
					self.min_size,
					self.max_size,
					self.fixed_size,
					self.margins,
					self.padding,
					self.helptext,
					self.position,
					self.style,
					self.hexpand,
					self.vexpand,
					self.font,
					self.base_color,
					self.background_color,
					self.foreground_color,
					self.selection_color,
					self.border_color,
					self.outline_color,
					self.border_size,
					self.outline_size,
					self.position_technique,
					self.is_focusable,
					self.comment,
					self.background_image,
					self.opaque,
					self.layout,
					self.spacing,
					self.uniform_size,
					None,
					self.title,
					self.titlebar_height,
					self.movable,
					self.resizable,
					self.top_resizable,
					self.right_resizable,
					self.bottom_resizable,
					self.left_resizable,
					self.shove,
					self.cursors,
					self.active,
					self.side,
					self.highlight_color)
		
		dockAreaClone.addChildren(self._cloneChildren(prefix))		     
		return dockAreaClone

	def addChild(self, widget):
		super(DockArea,self).addChild(widget)
		widget.docked = True

	def _getActiveDockArea(self): return self.real_widget.isActiveDockArea()
	def _setActiveDockArea(self, active): self.real_widget.setActiveDockArea(active)
	active = property(_getActiveDockArea, _setActiveDockArea)
	
	def isTopSide(self):
		return self.real_widget.isTopSide()
	
	def setTopSide(self, side):
		self.real_widget.setTopSide(side)
		
	def isRightSide(self):
		return self.real_widget.isRightSide()
	
	def setRightSide(self, side):
		self.real_widget.setRightSide(side)

	def isBottomSide(self):
		return self.real_widget.isBottomSide()
	
	def setBottomSide(self, side):
		self.real_widget.setBottomSide(side)

	def isLeftSide(self):
		return self.real_widget.isLeftSide()
	
	def setLeftSide(self, side):
		self.real_widget.setLeftSide(side)
		
	def _getSide(self):
		if self.real_widget.isTopSide():
			return 'top'
		elif self.real_widget.isRightSide():
			return 'right'
		elif self.real_widget.isBottomSide():
			return 'bottom'
		elif self.real_widget.isLeftSide():
			return 'left'
		return ''
	def _setSide(self, side):
		self.setTopSide(False)
		self.setRightSide(False)
		self.setBottomSide(False)
		self.setLeftSide(False)
		if side == 'top':
			self.setTopSide(True)
		elif side == 'right':
			self.setRightSide(True)
		elif side == 'bottom':
			self.setBottomSide(True)
		elif side == 'left':
			self.setLeftSide(True)
	side = property(_getSide, _setSide)

	highlight_color = ColorProperty("HighlightColor")



# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2013 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

from xml.sax import ContentHandler, parse, parseString, ErrorHandler


class FontFileParser(ContentHandler):
	def __init__(self):
		ContentHandler.__init__(self)
		self._fonts = {}

	def startElement(self, tag, rawAttributes):
		attributes = dict([(name, rawAttributes.getValue(name)) for name in rawAttributes.getNames()])
		if tag.lower() == "font":
			if not "name" in attributes:
				raise ValueError("Fonts must have a name")
			name = attributes["name"]
			del attributes["name"]
			self._fonts[name] = attributes

	def get(self, font, attribute, default=None):
		return self._fonts.get(font, {}).get(attribute, default)

	def fonts(self):
		return iter(self._fonts.keys())

	def parse(self, filename, errorHandler=ErrorHandler()):
		parse(filename, self, errorHandler)

	def parseString(self, string, errorHandler=ErrorHandler()):
		parseString(string, self, errorHandler)



# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2019 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

"""\
Pythonize FIFE

Import this extension to get a more
pythonic interface to FIFE.

Currently it implements the following
conveniences:

  - FIFE Exceptions print their message.
  - Automatic property generation for:
    - fife.Engine
    - fife.Instance
    - fife.Image
    - fife.Animation
    - fife.Point
    - fife.Rect
"""
from __future__ import print_function

from builtins import map
from fife import fife
from fife import fifechan
import re

__all__ = ()

fife.Exception.__str__ = fife.Exception.what
def _Color2Str(c):
	return 'Color(%s)' % ','.join(map(str,(c.r,c.g,c.b,c.a)))
fifechan.Color.__str__ = _Color2Str

classes = [ fife.Engine, fife.Instance, fife.Point, fife.Rect, fife.Image, fife.Animation,
 fife.RenderBackend, fife.Event, fife.Command, fifechan.Container ]

def createProperties():
	""" Autocreate properties for getXYZ/setXYZ functions.
	"""
	try:
		import inspect
		getargspec = inspect.getargspec
	except ImportError:
		print("Pythonize: inspect not available - properties are generated with dummy argspec.")
		getargspec = lambda func : ([],'args',None,None)

	def isSimpleGetter(func):
		if not callable(func):
			return False
		try:
			argspec = getargspec(func)
			return not (argspec[0] or [s for s in argspec[2:] if s])
		except TypeError as e:
			#print func, e
			return False

	def createNames(name):
		for prefix in ('get', 'is', 'are'):
			if name.startswith(prefix):
				new_name = name[len(prefix):]
				break
		settername   = 'set' + new_name
		propertyname = new_name[0].lower() + new_name[1:]
		return settername, propertyname

	getter = re.compile(r"^(get|are|is)[A-Z]")
	for class_ in classes:
		methods = [(name,attr) for name,attr in list(class_.__dict__.items())
		                       if isSimpleGetter(attr) ]
		setmethods = [(name,attr) for name,attr in list(class_.__dict__.items()) if callable(attr)]
		getters = []
		for name,method in methods:
			if getter.match(name):
				getters.append((name,method))
				settername, propertyname = createNames(name)
				setter = dict(setmethods).get(settername,None)
				#print name, settername, "--->",propertyname,'(',method,',',setter,')'
				setattr(class_,propertyname,property(method,setter))
		if not getters: continue

		# We need to override the swig setattr function
		# to get properties to work.
		class_._property_names = set([name for name,method in getters])
		def _setattr_wrapper_(self,*args):
			if name in class_._property_names:
				object.__setattr__(self,*args)
			else:
				class_.__setattr__(self,*args)
		class_.__setattr__ = _setattr_wrapper_

createProperties()



# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2019 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

"""\
PyChan event handling (internal).
=================================

Users shouldn't need to use this module directly.
L{widgets.Widget.capture} and L{widgets.Widget.mapEvents} provide
a convenient API to capture events.

Nevertheless to understand how its supposed to work
take a look at L{EventMapper} and L{EventListenerBase}

Event callbacks
---------------

You can either write callbacks yourself or
use L{tools.callbackWithArguments} or L{tools.attrSetCallback}
to generate suitable callbacks.

Here's an example callback::
   def dumpEventInfo(event=0,widget=0):
	  print widget, " received the event ", event

Note the signature - C{event} and C{widget} are keyword
arguments passed to the callback. If doesn't accept either
C{event} or C{widget} as argument, these are not passed.

This way a simple function which ignores C{event} or C{widget}
can be used, while they are available if needed.

Currently only one callback can be set per event. In case
you don't want to write your own callback that dispatches
to different callbacks you can use L{tools.chainCallbacks}.

Available Events
----------------

"""
from __future__ import print_function
from __future__ import absolute_import

from builtins import str
from builtins import range
from builtins import object
from future.utils import itervalues
from .compat import fifechan

from . import exceptions
from .internal import get_manager
from . import tools
import traceback
import weakref
from fife.extensions.fife_timer import Timer
from fife.extensions.pychan.tools import callbackWithArguments as cbwa

EVENTS = [
	"mouseEntered",
	"mouseExited",
	"mousePressed",
	"mouseReleased",
	"mouseClicked",
	"mouseMoved",
	"mouseWheelMovedUp",
	"mouseWheelMovedDown",
	"mouseWheelMovedRight",
	"mouseWheelMovedLeft",
	"mouseDragged",
	"action",
	"keyPressed",
	"keyReleased",
	"widgetResized",
	"widgetMoved",
	"widgetHidden",
	"widgetShown",
	"ancestorMoved",
	"ancestorHidden",
	"ancestorShown"
]

# Add the EVENTS to the docs.
__doc__ += "".join([" - %s\n" % event for event in EVENTS])

# The line before seems to leak the variable event into the global namespace ... remove that!
# This is a python problem, addressed in python3
try: del event
except:pass

MOUSE_EVENT, KEY_EVENT, ACTION_EVENT, WIDGET_EVENT = list(range(4))
def getEventType(name):
	if "mouse" in name:
		return MOUSE_EVENT
	if "key" in name:
		return KEY_EVENT
	if ("widget" in name) or ("ancestor" in name):
		return WIDGET_EVENT
	return ACTION_EVENT


CALLBACK_NONE_MESSAGE = """\
You passed None as parameter to %s.capture, which would normally remove a mapped event.
But there was no event mapped. Did you accidently call a function instead of passing it?
"""

class EventListenerBase(object):
	"""
	Redirector for event callbacks.
	Use *only* from L{EventMapper}.

	This class uses the SWIG director feature - overriden
	virtual methods are called from C++ to - listen to
	Fifechan events.

	"""
	def __init__(self):
		super(EventListenerBase,self).__init__()
		self.events = {}
		self.indent = 0
		self.debug = get_manager().debug
		self.is_attached = False

		# Enables event redirection to next pump cycle by default
		self._redirect = True
		self._timers = []
		self._deadtimers = []

	def attach(self,widget):
		"""
		Start receiving events.
		No need to call this manually.
		"""

		if self.is_attached:
			return
		if not self.events:
			return
		if self.debug: print("Attach:",self)
		self.doAttach(widget.real_widget)
		self.widget_ref = weakref.ref(widget)
		self.is_attached = True

	def detach(self):
		"""
		Stop receiving events.
		No need to call this manually.
		"""
		if not self.is_attached:
			return
		if self.debug: print("Detach:",self)
		self.is_attached = False

	def setRedirection(self, redirect):
		"""
		If enabled, the events are redirected to the next
		engine pump cycle. Otherwise the exectution is on
		the same engine cycle.
		"""
		self._redirect = redirect

	def _redirectEvent(self,name,event):
		self.indent += 4
		try:
			event = self.translateEvent(getEventType(name), event)
			if name in self.events:
				if self.debug: print("-"*self.indent, name)
				for f in itervalues(self.events[name]):
					if not self._redirect:
						f(event)
						continue

					def delayed_f(timer, f=f): # bind f during loop
						n_timer = timer()
						f( event )

						#FIXME: figure out a way to get rid of the dead timer list
						del self._deadtimers[:]

						if n_timer in self._timers:
							self._deadtimers.append(n_timer)
							self._timers.remove(n_timer)


					timer = Timer(repeat=1)
					timer._callback = cbwa(delayed_f, weakref.ref(timer))
					timer.start()

					self._timers.append(timer)

		except:
			print(name, repr(event))
			traceback.print_exc()
			raise

		finally:
			self.indent -= 4

	def translateEvent(self,event_type,event):
		if event_type == MOUSE_EVENT:
			return get_manager().hook.translate_mouse_event(event)
		if event_type == KEY_EVENT:
			return get_manager().hook.translate_key_event(event)
		return event

class _ActionEventListener(EventListenerBase,fifechan.ActionListener):
	def __init__(self):super(_ActionEventListener,self).__init__()
	def doAttach(self,real_widget): real_widget.addActionListener(self)
	def doDetach(self,real_widget): real_widget.removeActionListener(self)

	def action(self,e): self._redirectEvent("action",e)

class _MouseEventListener(EventListenerBase,fifechan.MouseListener):
	def __init__(self):super(_MouseEventListener,self).__init__()
	def doAttach(self,real_widget): real_widget.addMouseListener(self)
	def doDetach(self,real_widget): real_widget.removeMouseListener(self)

	def mouseEntered(self,e): self._redirectEvent("mouseEntered",e)
	def mouseExited(self,e): self._redirectEvent("mouseExited",e)
	def mousePressed(self,e): self._redirectEvent("mousePressed",e)
	def mouseReleased(self,e): self._redirectEvent("mouseReleased",e)
	def mouseClicked(self,e): self._redirectEvent("mouseClicked",e)
	def mouseMoved(self,e): self._redirectEvent("mouseMoved",e)
	def mouseWheelMovedUp(self,e): self._redirectEvent("mouseWheelMovedUp",e)
	def mouseWheelMovedDown(self,e): self._redirectEvent("mouseWheelMovedDown",e)
	def mouseWheelMovedRight(self,e): self._redirectEvent("mouseWheelMovedRight",e)
	def mouseWheelMovedLeft(self,e): self._redirectEvent("mouseWheelMovedLeft",e)
	def mouseDragged(self,e): self._redirectEvent("mouseDragged",e)

class _KeyEventListener(EventListenerBase,fifechan.KeyListener):
	def __init__(self):super(_KeyEventListener,self).__init__()
	def doAttach(self,real_widget): real_widget.addKeyListener(self)
	def doDetach(self,real_widget): real_widget.removeKeyListener(self)

	def keyPressed(self,e): self._redirectEvent("keyPressed",e)
	def keyReleased(self,e): self._redirectEvent("keyReleased",e)

class _WidgetEventListener(EventListenerBase, fifechan.WidgetListener):
	def __init__(self):super(_WidgetEventListener, self).__init__()
	def doAttach(self,real_widget): real_widget.addWidgetListener(self)
	def doDetach(self,real_widget): real_widget.removeWidgetListener(self)

	def widgetResized(self, e): self._redirectEvent("widgetResized",e)
	def widgetMoved(self, e): self._redirectEvent("widgetMoved",e)
	def widgetHidden(self, e): self._redirectEvent("widgetHidden",e)
	def widgetShown(self, e): self._redirectEvent("widgetShown",e)
	def ancestorMoved(self, e): self._redirectEvent("ancestorMoved",e)
	def ancestorHidden(self, e): self._redirectEvent("ancestorHidden",e)
	def ancestorShown(self, e): self._redirectEvent("ancestorShown",e)


class EventMapper(object):
	"""
	Handles events and callbacks for L{widgets.Widget}
	and derived classes.

	Every PyChan widget has an L{EventMapper} instance
	as attribute B{event_mapper}.

	This instance handles all necessary house-keeping.
	Such an event mapper can be either C{attached} or
	C{detached}. In its attached state an L{EventListenerBase}
	is added to the Fifechan widget and will redirect
	the events to the callbacks.

	In its detached state no events are received from the
	real Fifechan widget.

	The event mapper starts in the detached state.
	When a new event is captured the mapper attaches itself
	automatically. The widget doesn't need to handle that.
	"""
	def __init__(self,widget):
		super(EventMapper,self).__init__()
		self.widget_ref = weakref.ref(widget)
		self.callbacks = {}
		self.listener = {
			KEY_EVENT    : _KeyEventListener(),
			ACTION_EVENT : _ActionEventListener(),
			MOUSE_EVENT  : _MouseEventListener(),
			WIDGET_EVENT : _WidgetEventListener()
		}
		self.is_attached = False
		self.debug = get_manager().debug

	def __repr__(self):
		return "EventMapper(%s)" % repr(self.widget_ref())

	def attach(self):
		for listener in list(self.listener.values()):
			listener.attach()

	def detach(self):
		for listener in list(self.listener.values()):
			listener.detach()


	def capture(self,event_name,callback,group_name):
		if event_name not in EVENTS:
			raise exceptions.RuntimeError("Unknown eventname: " + event_name)

		if callback is None:
			if self.isCaptured(event_name,group_name):
				self.removeEvent(event_name,group_name)
			elif self.debug:
				print(CALLBACK_NONE_MESSAGE % str(self.widget_ref()))
			return
		self.addEvent(event_name,callback,group_name)

	def isCaptured(self,event_name,group_name="default"):
		return ("%s/%s" % (event_name,group_name)) in self.getCapturedEvents()

	def getCapturedEvents(self):
		events = []
		for event_type, listener in list(self.listener.items()):
			for event_name, group in list(listener.events.items()):
				for group_name in list(group.keys()):
					events.append( "%s/%s" % (event_name, group_name) )
		return events

	def getListener(self,event_name):
		return self.listener[getEventType(event_name)]

	def removeEvent(self,event_name,group_name):
		listener = self.getListener(event_name)
		del listener.events[event_name][group_name]

		if not listener.events[event_name]:
			del listener.events[event_name]
		if not listener.events:
			listener.detach()

		del self.callbacks[group_name][event_name]
		if len(self.callbacks[group_name]) <= 0:
			del self.callbacks[group_name]

	def addEvent(self,event_name,callback,group_name):
		if not callable(callback):
			raise RuntimeError("An event callback must be either a callable or None - not %s" % repr(callback))
		# The closure self needs to keep a weak ref.
		# Otherwise the GC has problems.
		self_ref = weakref.ref(self)

		# Set up callback dictionary. This should fix some GC issues
		if group_name not in self.callbacks:
			self.callbacks[group_name] = {}

		if event_name not in self.callbacks[group_name]:
			self.callbacks[group_name][event_name] = {}

		self.callbacks[group_name][event_name] = callback

		def captured_f(event):
			if self_ref() is not None:
				tools.applyOnlySuitable(self_ref().callbacks[group_name][event_name],event=event,widget=self_ref().widget_ref())

		listener = self.getListener(event_name)

		if event_name not in listener.events:
			listener.events[event_name] = {group_name : captured_f}
		else:
			listener.events[event_name][group_name] = captured_f
		listener.attach(self.widget_ref())


def splitEventDescriptor(name):
	""" Utility function to split "widgetName/eventName" descriptions into tuples. """
	L = name.split("/")
	if len(L) not in (1,2,3):
		raise exceptions.RuntimeError("Invalid widgetname / eventname combination: " + name)
	if len(L) == 1:
		L = L[0],"action"
	elif L[1] not in EVENTS:
		raise exceptions.RuntimeError("Unknown event name: " + name)
	if len(L) == 2:
		L = L[0],L[1],"default"
	return L




# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2013 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

from __future__ import absolute_import
import weakref
from fife import fife
from fife import fifechan

from fife.extensions.pychan.attrs import BoolAttr
from .common import get_manager

from .resizablewindow import ResizableWindow
from .dockarea import DockArea
from .containers import Container

class Panel(ResizableWindow):
	"""
	The Panel class can be docked or undocked from Dock Areas.
	If the Panel is added to a DockArea e.g. by XML loading, then it will automatically docked.
	By default undock will add the Panel to the parent of the DockArea, that can also be the top widget.

	New Attributes
	==============

	  - dockable: If true, the Panel can be docked/undocked to DockAreas.
	"""
	
	ATTRIBUTES = ResizableWindow.ATTRIBUTES + [ BoolAttr('dockable'),
												]


	DEFAULT_DOCKABLE = True


	def __init__(self, 
				 parent = None, 
				 name = None,
				 size = None,
				 min_size = None, 
				 max_size = None,
				 fixed_size = None,
				 margins = None,
				 padding = None,
				 helptext = None, 
				 position = None, 
				 style = None, 
				 hexpand = None,
				 vexpand = None,
				 font = None,
				 base_color = None,
				 background_color = None,
				 foreground_color = None,
				 selection_color = None,
				 border_color = None,
				 outline_color = None,
				 border_size = None,
				 outline_size = None,
				 position_technique = None,
				 is_focusable = None,
				 comment = None,
				 background_image = None,
				 opaque = None,
				 _real_widget = None,
				 title = None,
				 titlebar_height = None,
				 movable = None,
				 resizable = None,
				 top_resizable = None,
				 right_resizable = None,
				 bottom_resizable = None,
				 left_resizable = None,
				 shove = None,
				 cursors = None,
				 dockable = None):
		
		if _real_widget is None: _real_widget = fifechan.Panel()
			
		super(Panel,self).__init__(parent=parent,
								   name=name,
								   size=size,
								   min_size=min_size,
								   max_size=max_size,
								   fixed_size=fixed_size,
								   margins=margins,
								   padding=padding,
								   helptext=helptext,
								   position=position,
								   style=style,
								   hexpand=hexpand,
								   vexpand=vexpand,
								   font=font,
								   base_color=base_color,
								   background_color=background_color,
								   foreground_color=foreground_color,
								   selection_color=selection_color,
								   border_color=border_color,
								   outline_color=outline_color,
								   border_size=border_size,
								   outline_size=outline_size,
								   position_technique=position_technique,
								   is_focusable=is_focusable,
								   comment=comment,
								   background_image=background_image,
								   opaque=opaque,
								   _real_widget=_real_widget,
								   title=title,
								   titlebar_height=titlebar_height,
								   movable=movable,
								   resizable=resizable,
								   top_resizable=top_resizable,
								   right_resizable=right_resizable,
								   bottom_resizable=bottom_resizable,
								   left_resizable=left_resizable,
								   shove=shove,
								   cursors=cursors)

		if dockable is not None: self.dockable = dockable
		else: self.dockable = self.DEFAULT_DOCKABLE
		
		self._foundDockArea = None
		
		self._barPressedLeft = False
		self._barPressedRight = False
		self._barReleasedLeft = False
		self._barReleasedRight = False
		
		self.capture(self.mousePressed, "mousePressed", "Panel")
		self.capture(self.mouseReleased, "mouseReleased", "Panel")
		self.capture(self.mouseDragged, "mouseDragged", "Panel")

	def clone(self, prefix):
		panelClone = Panel(None, 
					self._createNameWithPrefix(prefix),
					self.size,
					self.min_size, 
					self.max_size,
					self.fixed_size,
					self.margins,
					self.padding,
					self.helptext, 
					self.position, 
					self.style, 
					self.hexpand,
					self.vexpand,
					self.font,
					self.base_color,
					self.background_color,
					self.foreground_color,
					self.selection_color,
					self.border_color,
					self.outline_color,
					self.border_size,
					self.outline_size,
					self.position_technique,
					self.is_focusable,
					self.comment,
					self.background_image,
					self.opaque,
					None,
					self.title,
					self.titlebar_height,
					self.movable,
					self.resizable,
					self.top_resizable,
					self.right_resizable,
					self.bottom_resizable,
					self.left_resizable,
					self.shove,
					self.cursors,
					self.dockable)
		
		panelClone.addChildren(self._cloneChildren(prefix))		     
		return panelClone

	def _getDocked(self): return self.real_widget.isDocked()
	def _setDocked(self, docked): self.real_widget.setDocked(docked)
	docked = property(_getDocked, _setDocked)

	def _getDockable(self): return self.real_widget.isDockable()
	def _setDockable(self, dockable): self.real_widget.setDockable(dockable)
	dockable = property(_getDockable, _setDockable)
	
	def getDockArea(self):
		if not self.docked:
			dockAreas = []
			# all top widgets are used for the search
			if not self.parent:
				topWidgets = get_manager().allTopHierachyWidgets
				for t in topWidgets:
					dockAreas.extend(t.findChildren(__class__=DockArea))
			else:
				# only all childs are used for the search
				dockAreas = self.parent.findChildren(__class__=DockArea)
			# reverse order so inner/deeper Areas are preferred
			dockAreas.reverse()
			# try to find an intersecting and active DockArea
			dim = fife.Rect(0, 0, self.width, self.height)
			dim.x, dim.y = self.getAbsolutePos()
			for d in dockAreas:
				if d.real_widget.isActiveDockArea():
					ddim = fife.Rect(0, 0, d.width, d.height)
					ddim.x, ddim.y = d.getAbsolutePos()
					if dim.intersects(ddim):
						return d
			return None
		else:
			return self.parent

	def afterDock(self):
		""" called after dockTo() process if the panel is docked
		
			overwrite this to e.g. keep a record on where the widget
			was last docked
		"""
		pass
		
	def afterUndock(self):
		""" called after undockTo() process if the panel is undocked 
		
			overwrite and implement this in your Panel instance
			to e.g. restore a particular default position
			(otherwise the panel would re-appear on the center)
		"""
		pass
	
	def dockTo(self, widget):
		# Dock the Panel to the given widget.
		if not self.docked and widget is not self.parent and self.dockable:
			widget.real_widget.setHighlighted(False)
			# map coordinates to new parent and remove it from old parent
			if self.parent:
				self.x = (self.x // (self.parent.width // 100)) * (widget.width // 100)
				self.y = (self.y // (self.parent.height // 100)) * (widget.height // 100)
				self.parent.removeChild(self)
			else:
				self.x = (self.x // (get_manager().hook.screen_width // 100)) * (widget.width // 100)
				self.y = (self.y // (get_manager().hook.screen_height // 100)) * (widget.height // 100)
				get_manager().removeTopWidget(self)
			# dock it to new parent
			widget.addChild(self)
			self.docked = True
			self.afterDock()

	def undockTo(self, widget):
		# Undock the Panel to the given widget.
		if self.docked and widget is not self.parent and self.dockable:
			self.parent.removeChild(self)
			# undock to main gui
			if widget is None:
				get_manager().addTopWidget(self)
			else:
				widget.addChild(self)
			self.docked = False
			self.afterUndock()
		
	def mousePressed(self, event):
		h = self.real_widget.getBorderSize() + self.real_widget.getPaddingTop() + self.real_widget.getTitleBarHeight()
		self._barPressedLeft = event.getButton() == 1 and event.getY() <= h and event.getY() > self.real_widget.getResizableBorderDistance()
		self._barPressedRight = event.getButton() == 2 and event.getY() <= h and event.getY() > self.real_widget.getResizableBorderDistance()
		self._barReleasedLeft = False
		self._barReleasedRight = False

	def mouseReleased(self, event):
		h = self.real_widget.getBorderSize() + self.real_widget.getPaddingTop() + self.real_widget.getTitleBarHeight()
		self._barReleasedLeft = event.getButton() == 1 and event.getY() <= h and event.getY() > self.real_widget.getResizableBorderDistance()
		self._barReleasedRight = event.getButton() == 2 and event.getY() <= h and event.getY() > self.real_widget.getResizableBorderDistance()
		releasedLeft = self._barPressedLeft and self._barReleasedLeft
		releasedRight = self._barPressedRight and self._barReleasedRight
		self._barPressedLeft = False
		self._barPressedRight = False
		if releasedLeft and self._foundDockArea and not self.docked:
			self.dockTo(self._foundDockArea())
		elif releasedRight and self.docked:
			# by default it undocks to the parent of the DockArea
			if self.parent.parent:
				newParent = self.parent.parent
				self.undockTo(newParent)
			else:
				self.undockTo(None)

	def mouseDragged(self, event):
		# disable highlighting
		if self._foundDockArea is not None:
			self._foundDockArea().real_widget.setHighlighted(False)
			self._foundDockArea = None

		if not self.docked and self._barPressedLeft:
			dock = self.getDockArea()
			# enable highlighting for dock area
			if dock is not None and dock.real_widget.isActiveDockArea():
				self._foundDockArea = weakref.ref(dock)
				self._foundDockArea().real_widget.setHighlighted(True)
